"""
This module defines an :py:class:`aiohttp.ClientSession` adapter that
returns :py:class:`twisted.internet.defer.Deferred` responses.
"""

# Third party imports
try:
    from twisted.internet import threads
except ImportError:  # pragma: no cover
    threads = None

# Local imports
from uplink.clients import interfaces, io, register


class TwistedClient(interfaces.HttpClientAdapter):
    """
    Client that returns :py:class:`twisted.internet.defer.Deferred`
    responses.

    Note:
        This client is an optional feature and requires the :py:mod:`twisted`
        package. For example, here's how to install this extra using pip::

            $ pip install uplink[twisted]

    Args:
        session (:py:class:`requests.Session`, optional): The session
            that should handle sending requests. If this argument is
            omitted or set to :py:obj:`None`, a new session will be
            created.
    """

    def __init__(self, session=None):
        if threads is None:
            raise NotImplementedError("twisted is not installed.")
        self._proxy = register.get_client(session)

    @property
    def exceptions(self):
        return self._proxy.exceptions

    @staticmethod
    def io():
        return io.TwistedStrategy()

    def apply_callback(self, callback, response):
        return threads.deferToThread(
            self._proxy.apply_callback, callback, response
        )

    def send(self, request):
        return threads.deferToThread(self._proxy.send, request)



"""This module implements the auth layer."""

# Third-party imports
from requests import auth

# Local imports
from uplink import utils
from uplink.compat import abc

__all__ = [
    "ApiTokenParam",
    "ApiTokenHeader",
    "BasicAuth",
    "ProxyAuth",
    "BearerToken",
    "MultiAuth",
]


def get_auth(auth_object=None):
    if auth_object is None:
        return utils.no_op
    elif isinstance(auth_object, abc.Iterable):
        return BasicAuth(*auth_object)
    elif callable(auth_object):
        return auth_object
    else:
        raise ValueError("Invalid authentication strategy: %s" % auth_object)


class ApiTokenParam(object):
    """
    Authorizes requests using a token or key in a query parameter.

    Users may use this directly, or API library authors may subclass this
    to predefine the query parameter name to use. If supplying query parameter
    name on a subclass, define the ``_param`` property or attribute and
    override ``__init__()`` without using ``super()``.

    .. code-block:: python

        # direct use
        token_param = ApiTokenParam(QUERY_PARAM_NAME, TOKEN)
        api_consumer = SomeApiConsumerClass(BASE_URL, auth=token_param)

        # subclass in API library
        class ExampleApiTokenParam(ApiTokenParam):
            _param = "api-token"
            def __init__(self, token):
                self._param_value = token

        # using the subclass
        token_param = ExampleApiTokenParam(TOKEN)
        api_consumer = SomeApiConsumerClass(BASE_URL, auth=token_param)
    """

    def __init__(self, param, token):
        self._param = param
        self._param_value = token

    def __call__(self, request_builder):
        request_builder.info["params"][self._param] = self._param_value


class ApiTokenHeader(object):
    """
    Authorizes requests using a token or key in a header.
    Users should subclass this class to define which header is the token header.
    The subclass may also, optionally, define a token prefix (such as in BearerToken)

    Users may use this directly, or API library authors may subclass this
    to predefine the header name or a prefix to use. If supplying header name or prefix
    in a subclass, define the ``_header`` and/or ``_prefix`` properties or attributes
    and override ``__init__()`` without using ``super()``.

    .. code-block:: python

        # direct use
        token_header = ApiTokenHeader(HEADER_NAME, TOKEN)
        api_consumer = SomeApiConsumerClass(BASE_URL, auth=token_header)

        # subclass in API library with a prefix
        class ExampleApiTokenHeader(ApiTokenHeader):
            _header = "X-Api-Token"
            def __init__(self, token):
                self._token = token

        # subclass in API library with a prefix
        class ExampleApiTokenHeader(ApiTokenHeader):
            _header = "X-App-Id"
            _prefix = "APP"
            def __init__(self, token):
                self._token = token

        # using the subclass
        token_header = ExampleApiTokenHeader(TOKEN)
        api_consumer = SomeApiConsumerClass(BASE_URL, auth=token_header)
    """

    _header = None
    _prefix = None

    def __init__(self, header, token, prefix=None):
        self._header = header
        self._prefix = prefix
        self._token = token

    @property
    def _header_value(self):
        if self._prefix:
            return "%s %s" % (self._prefix, self._token)
        else:
            return self._token

    def __call__(self, request_builder):
        request_builder.info["headers"][self._header] = self._header_value


class BasicAuth(ApiTokenHeader):
    """
    Authorizes requests using HTTP Basic Authentication.

    There are two ways to use BasicAuth with a Consumer:

    .. code-block:: python

        # implicit BasicAuth
        github = Github(BASE_URL, auth=(USER, PASS))

        # explicit BasicAuth
        github = GitHub(BASE_URL, auth=BasicAuth(USER, PASS))
    """

    _header = "Authorization"

    def __init__(self, username, password):
        self._username = username
        self._password = password

    @property
    def _header_value(self):
        return auth._basic_auth_str(self._username, self._password)


class ProxyAuth(BasicAuth):
    """
    Authorizes requests with an intermediate HTTP proxy.

    If both API auth and intermediate Proxy auth are required,
    wrap ProxyAuth in MultiAuth:

    .. code-block:: python

        # only ProxyAuth
        github = Github(BASE_URL, auth=ProxyAuth(PROXYUSER, PROXYPASS))

        # both BasicAuth and ProxyAuth
        auth_methods = MultiAuth(
            BasicAuth(USER, PASS),
            ProxyAuth(PROXYUSER, PROXYPASS)
        )
        github = GitHub(BASE_URL, auth=auth_methods)
    """

    _header = "Proxy-Authorization"


class BearerToken(ApiTokenHeader):
    """
    Authorizes requests using a Bearer Token.

    .. code-block:: python

        token = something_like_oauth_that_returns_a_token()
        bearer = BearerToken(token)
        api_consumer = SomeApiConsumerClass(BASE_URL, auth=bearer)
    """

    _header = "Authorization"
    _prefix = "Bearer"

    def __init__(self, token):
        self._token = token


class MultiAuth(object):
    """
    Authorizes requests using multiple auth methods at the same time.

    This is useful for API users to supply both API credentials and
    intermediary credentials (such as for a proxy).

    .. code-block:: python

        auth_methods = MultiAuth(
            BasicAuth(USER, PASS),
            ProxyAuth(PROXY_USER, PROXY_PASS)
        )
        api_consumer = SomeApiConsumerClass(BASE_URL, auth=auth_methods)

    This may also be used if an API requires multiple Auth Tokens.

    .. code-block:: python

        auth_methods = MultiAuth(
            BearerToken(API_TOKEN),
            ApiTokenParam(QUERY_PARAMETER_NAME, QUERY_PARAMETER_VALUE),
            ApiTokenHeader(API_HEADER_NAME, API_TOKEN_2)
        )
        api_consumer = SomeApiConsumerClass(BASE_URL, auth=auth_methods)

    API library authors may find it more helpful to treat MultiAuth as
    a list using ``append`` or ``extend`` to add aditional auth methods.

    .. code-block:: python

        auth_methods = MultiAuth()

        auth_methods.append(BearerToken(API_TOKEN))
        auth_methods.extend([
            ApiTokenParam(QUERY_PARAMETER_NAME, QUERY_PARAMETER_VALUE),
            ApiTokenHeader(API_HEADER_NAME, API_TOKEN_2)
        ])
        api_consumer = SomeApiConsumerClass(BASE_URL, auth=auth_methods)

        # looping over contained auth methods is also supported
        for method in auth_methods:
            print(method.__class__.__name__)
    """

    def __init__(self, *auth_methods):
        self._auth_methods = [
            get_auth(auth_method) for auth_method in auth_methods
        ]

    def __call__(self, request_builder):
        for auth_method in self._auth_methods:
            auth_method(request_builder)

    def __getitem__(self, index):
        return self._auth_methods[index]

    def __len__(self):
        return len(self._auth_methods)

    def append(self, auth_method):
        self._auth_methods.append(get_auth(auth_method))

    def extend(self, auth_methods):
        self._auth_methods.extend(
            [get_auth(auth_method) for auth_method in auth_methods]
        )



"""
This package defines an adapter layer for writing wrappers of existing
HTTP clients (Requests, aiohttp, etc.) so they can handle requests built by
Uplink's high-level, declarative API.

We refer to this layer as the backend, as these adapters handle the
actual HTTP client logic (i.e., making a request to a server).

Todo:
    At some point, we may want to expose this layer to the user, so
    they can create custom adapters.
"""
# Local imports
from uplink import utils
from uplink.clients import interfaces, register
from uplink.clients.register import DEFAULT_CLIENT, get_client
from uplink.clients.requests_ import RequestsClient
from uplink.clients.twisted_ import TwistedClient


@register.handler
def _client_class_handler(key):
    if utils.is_subclass(key, interfaces.HttpClientAdapter):
        return key()


try:
    from uplink.clients.aiohttp_ import AiohttpClient
except (ImportError, SyntaxError):  # pragma: no cover

    class AiohttpClient(interfaces.HttpClientAdapter):
        def __init__(self, *args, **kwargs):
            raise NotImplementedError(
                "Failed to load `aiohttp` client: you may be using a version "
                "of Python below 3.3. `aiohttp` requires Python 3.4+."
            )


__all__ = [
    "RequestsClient",
    "AiohttpClient",
    "TwistedClient",
    "DEFAULT_CLIENT",
    "get_client",
]

register.set_default_client(RequestsClient)



# Third-party imports
import pytest

# Local imports
from uplink import commands, converters, arguments, utils


class TestHttpMethodFactory(object):
    def test_call_as_decorator_with_no_args(self):
        @commands.HttpMethodFactory(None)
        def func():
            pass

        assert isinstance(func, commands.RequestDefinitionBuilder)

    def test_call_as_decorator_with_args(self):
        method_factory = commands.HttpMethodFactory(None)

        @method_factory(None)
        def func():
            pass

        assert isinstance(func, commands.RequestDefinitionBuilder)


class TestHttpMethod(object):
    def test_call(self, mocker, annotation_mock):
        # Setup
        def func():
            pass

        sig = utils.Signature(
            args=["self", "arg1", "arg2"],
            annotations={"arg1": annotation_mock},
            return_annotation=None,
        )
        mocker.patch("uplink.utils.get_arg_spec").return_value = sig

        http_method = commands.HttpMethod("METHOD", uri="/{hello}")
        builder = http_method(func)
        assert isinstance(builder, commands.RequestDefinitionBuilder)
        assert builder.method == "METHOD"
        assert list(builder.uri.remaining_variables) == ["hello"]

        missing_arguments = builder.argument_handler_builder.missing_arguments
        expected_missing = set(sig.args[1:]) - set(sig.annotations.keys())
        assert set(missing_arguments) == expected_missing

    def test_call_with_return_annotation(self, mocker):
        # Setup
        def func():
            pass

        sig = utils.Signature(
            args=[], annotations={}, return_annotation="return_annotation"
        )
        mocker.patch("uplink.utils.get_arg_spec").return_value = sig
        returns = mocker.patch("uplink.returns.schema")
        http_method = commands.HttpMethod("METHOD", uri="/{hello}")
        http_method(func)

        # Verify: build is wrapped with decorators.returns
        returns.assert_called_with(sig.return_annotation)

    def test_call_with_args(self, mocker, annotation_mock):
        # Setup
        def func():
            pass

        args = mocker.patch("uplink.decorators.args")

        # Verify: using sequence
        http_method = commands.HttpMethod(
            "METHOD", uri="/{hello}", args=(annotation_mock,)
        )
        http_method(func)
        args.assert_called_with(annotation_mock)

        # Verify: using mapping
        http_method = commands.HttpMethod(
            "METHOD", uri="/{hello}", args={"arg1": "value"}
        )
        http_method(func)
        args.assert_called_with(arg1="value")


class TestURIDefinitionBuilder(object):
    def test_is_static(self):
        assert not commands.URIDefinitionBuilder(None).is_static

    def test_is_dynamic_setter(self):
        uri = commands.URIDefinitionBuilder(None)
        assert not uri.is_dynamic
        uri.is_dynamic = True
        assert uri.is_dynamic

    def test_is_dynamic_setter_fails_when_is_static(self):
        uri = commands.URIDefinitionBuilder(True)
        assert uri.is_static
        with pytest.raises(ValueError):
            uri.is_dynamic = True

    def test_remaining_variables(self):
        uri = commands.URIDefinitionBuilder("/path/with/{variable}")
        assert uri.remaining_variables == set(["variable"])

    def test_add_variable(self):
        uri = commands.URIDefinitionBuilder("/path/with/{variable}")
        assert "variable" in uri.remaining_variables
        uri.add_variable("variable")
        assert "variable" not in uri.remaining_variables

    def test_add_variable_raise_error_when_name_is_not_in_static_path(self):
        uri = commands.URIDefinitionBuilder("/static/path")
        with pytest.raises(ValueError):
            uri.add_variable("variable")

    def test_build(self):
        uri = commands.URIDefinitionBuilder("/static/path")
        assert uri.build() == "/static/path"

    def test_build_fails_when_variable_remain_in_uri(self):
        uri = commands.URIDefinitionBuilder("/path/with/{variable}")
        with pytest.raises(commands.MissingUriVariables):
            uri.build()


class TestRequestDefinitionBuilder(object):
    def test_method_handler_builder_getter(
        self, annotation_handler_builder_mock
    ):
        builder = commands.RequestDefinitionBuilder(
            None,
            None,
            None,
            type(annotation_handler_builder_mock)(),
            annotation_handler_builder_mock,
        )
        assert builder.method_handler_builder is annotation_handler_builder_mock

    def test_build(self, mocker, annotation_handler_builder_mock):
        argument_handler_builder = type(annotation_handler_builder_mock)()
        method_handler_builder = annotation_handler_builder_mock
        uri_definition_builder = mocker.Mock(spec=commands.URIDefinitionBuilder)
        builder = commands.RequestDefinitionBuilder(
            None,
            "method",
            uri_definition_builder,
            argument_handler_builder,
            method_handler_builder,
        )
        definition = builder.build()
        assert isinstance(definition, commands.RequestDefinition)
        assert uri_definition_builder.build.called
        assert argument_handler_builder.build.called
        assert method_handler_builder.build.called

    def test_auto_fill_when_not_done(
        self, mocker, annotation_handler_builder_mock
    ):
        # Setup
        argument_handler_builder = mocker.Mock(
            stub=arguments.ArgumentAnnotationHandlerBuilder
        )
        method_handler_builder = annotation_handler_builder_mock
        uri_definition_builder = mocker.Mock(spec=commands.URIDefinitionBuilder)
        builder = commands.RequestDefinitionBuilder(
            None,
            "method",
            uri_definition_builder,
            argument_handler_builder,
            method_handler_builder,
        )

        # Setup success condition
        argument_handler_builder.is_done.return_value = False
        argument_handler_builder.missing_arguments = ["arg1"]
        uri_definition_builder.remaining_variables = ["arg1"]

        # Verify
        builder.build()
        argument_handler_builder.set_annotations.assert_called_with(
            {"arg1": arguments.Path}
        )

    def test_auto_fill_when_not_done_fails(
        self, mocker, annotation_handler_builder_mock
    ):
        # Setup
        argument_handler_builder = annotation_handler_builder_mock
        method_handler_builder = annotation_handler_builder_mock
        uri_definition_builder = mocker.Mock(spec=commands.URIDefinitionBuilder)
        builder = commands.RequestDefinitionBuilder(
            None,
            "method",
            uri_definition_builder,
            argument_handler_builder,
            method_handler_builder,
        )

        # Setup fail condition: Argument is missing annotation
        argument_handler_builder.is_done.return_value = False
        argument_handler_builder.missing_arguments = ["arg1"]
        uri_definition_builder.remaining_variables = []

        # Verify
        with pytest.raises(commands.MissingArgumentAnnotations):
            builder.build()


class TestRequestDefinition(object):
    def test_argument_annotations(self, annotation_handler_mock):
        annotation_handler_mock.annotations = ["arg1", "arg2"]
        definition = commands.RequestDefinition(
            None, None, None, annotation_handler_mock, None
        )
        assert list(definition.argument_annotations) == ["arg1", "arg2"]

    def test_method_annotations(self, annotation_handler_mock):
        annotation_handler_mock.annotations = ["arg1", "arg2"]
        definition = commands.RequestDefinition(
            None, None, None, None, annotation_handler_mock
        )
        assert list(definition.method_annotations) == ["arg1", "arg2"]

    def test_define_request(self, request_builder, mocker):
        method = "method"
        uri = "uri"
        definition = commands.RequestDefinition(
            method, uri, str, mocker.Mock(), mocker.Mock()
        )
        definition.define_request(request_builder, (), {})
        assert request_builder.method == method
        assert request_builder.relative_url == uri
        assert request_builder.return_type is str

    def test_make_converter_registry(self, annotation_handler_mock):
        definition = commands.RequestDefinition(
            "method",
            "uri",
            None,
            annotation_handler_mock,
            annotation_handler_mock,
        )
        annotation_handler_mock.annotations = ("annotation",)
        registry = definition.make_converter_registry(())
        assert isinstance(registry, converters.ConverterFactoryRegistry)



# Standard library imports
import contextlib

# Third-party imports
import pytest

# Local imports
from uplink.clients import (
    interfaces,
    requests_,
    twisted_,
    register,
    io,
)


@contextlib.contextmanager
def _patch(obj, attr, value):
    if obj is not None:
        old_value = getattr(obj, attr)
        setattr(obj, attr, value)
    yield
    if obj is not None:
        setattr(obj, attr, old_value)


def test_get_default_client_with_non_callable():
    # Setup
    old_default = register.get_default_client()
    register.set_default_client("client")
    default_client = register.get_default_client()
    register.set_default_client(old_default)

    # Verify: an object that is not callable should be returned as set.
    assert default_client == "client"


def test_get_client_with_http_client_adapter_subclass():
    class HttpClientAdapterMock(interfaces.HttpClientAdapter):
        pass

    client = register.get_client(HttpClientAdapterMock)
    assert isinstance(client, HttpClientAdapterMock)


def test_get_client_with_unrecognized_key():
    assert register.get_client("no client for this key") is None


class TestRequests(object):
    def test_get_client(self, mocker):
        import requests

        session_mock = mocker.Mock(spec=requests.Session)
        client = register.get_client(session_mock)
        assert isinstance(client, requests_.RequestsClient)

    def test_init_with_kwargs(self):
        session = requests_.RequestsClient._create_session(
            cred=("username", "password")
        )
        assert session.cred == ("username", "password")

    def test_client_send(self, mocker):
        # Setup
        import requests

        session_mock = mocker.Mock(spec=requests.Session)
        session_mock.request.return_value = "response"
        callback = mocker.stub()
        client = requests_.RequestsClient(session_mock)

        # Run
        response = client.send(("method", "url", {}))

        # Verify
        session_mock.request.assert_called_with(method="method", url="url")

        # Run callback
        client.apply_callback(callback, response)
        callback.assert_called_with(session_mock.request.return_value)

    def test_dont_close_provided_session(self, mocker):
        # Setup
        import requests
        import gc

        session_mock = mocker.Mock(spec=requests.Session)
        session_mock.request.return_value = "response"

        # Run
        client = requests_.RequestsClient(session_mock)
        client.send(("method", "url", {}))
        del client
        gc.collect()

        assert session_mock.close.call_count == 0

    def test_close_auto_generated_session(self, mocker):
        # Setup
        import requests
        import gc

        session_mock = mocker.Mock(spec=requests.Session)
        session_mock.request.return_value = "response"
        session_cls_mock = mocker.patch("requests.Session")
        session_cls_mock.return_value = session_mock

        # Run
        client = requests_.RequestsClient()
        client.send(("method", "url", {}))
        del client
        gc.collect()

        assert session_mock.close.call_count == 1

    def test_exceptions(self):
        import requests

        exceptions = requests_.RequestsClient.exceptions

        with pytest.raises(exceptions.BaseClientException):
            raise requests.RequestException()

        with pytest.raises(exceptions.BaseClientException):
            # Test polymorphism
            raise requests.exceptions.InvalidURL()

        with pytest.raises(exceptions.ConnectionError):
            raise requests.exceptions.ConnectionError()

        with pytest.raises(exceptions.ConnectionTimeout):
            raise requests.exceptions.ConnectTimeout()

        with pytest.raises(exceptions.ServerTimeout):
            raise requests.exceptions.ReadTimeout()

        with pytest.raises(exceptions.SSLError):
            raise requests.exceptions.SSLError()

        with pytest.raises(exceptions.InvalidURL):
            raise requests.exceptions.InvalidURL()

    def test_io(self):
        assert isinstance(requests_.RequestsClient.io(), io.BlockingStrategy)


class TestTwisted(object):
    def test_init_without_client(self):
        twisted = twisted_.TwistedClient()
        assert isinstance(twisted._proxy, requests_.RequestsClient)

    def test_create_requests_no_twisted(self, http_client_mock):
        with _patch(twisted_, "threads", None):
            with pytest.raises(NotImplementedError):
                twisted_.TwistedClient(http_client_mock)

    def test_client_send(self, mocker, http_client_mock):
        deferToThread = mocker.patch.object(twisted_.threads, "deferToThread")
        request = twisted_.TwistedClient(http_client_mock)
        request.send((1, 2, 3))
        deferToThread.assert_called_with(http_client_mock.send, (1, 2, 3))

    def test_client_callback(self, mocker, http_client_mock):
        # Setup
        callback = mocker.stub()
        deferred = mocker.Mock()
        deferToThread = mocker.patch.object(twisted_.threads, "deferToThread")
        deferToThread.return_value = deferred
        client = twisted_.TwistedClient(http_client_mock)

        # Run
        response = client.apply_callback(callback, 1)

        # Verify
        assert response is deferred
        deferToThread.assert_called_with(
            http_client_mock.apply_callback, callback, 1
        )

    def test_exceptions(self, http_client_mock):
        twisted_client = twisted_.TwistedClient(http_client_mock)
        assert http_client_mock.exceptions == twisted_client.exceptions

    def test_io(self):
        assert isinstance(twisted_.TwistedClient.io(), io.TwistedStrategy)



import sys

collect_ignore = []
if sys.version_info.major < 3:
    collect_ignore.extend(
        [
            "unit/test_aiohttp_client.py",
            "integration/test_handlers_aiohttp.py",
            "integration/test_retry_aiohttp.py",
        ]
    )



"""
This module implements the built-in class and method decorators and their
handling classes.
"""
# Standard library imports
import functools
import inspect

# Local imports
from uplink import arguments, helpers, hooks, interfaces, utils
from uplink.compat import abc

__all__ = [
    "headers",
    "params",
    "form_url_encoded",
    "multipart",
    "json",
    "timeout",
    "args",
    "response_handler",
    "error_handler",
    "inject",
]


class MethodAnnotationHandlerBuilder(interfaces.AnnotationHandlerBuilder):
    def __init__(self):
        self._class_annotations = list()
        self._method_annotations = list()

    def add_annotation(self, annotation, *args_, **kwargs):
        if kwargs.get("is_class", False):
            self._class_annotations.append(annotation)
        else:
            self._method_annotations.append(annotation)
        super(MethodAnnotationHandlerBuilder, self).add_annotation(annotation)
        return annotation

    def copy(self):
        clone = MethodAnnotationHandlerBuilder()
        clone._class_annotations = list(self._class_annotations)
        clone._method_annotations = list(self._method_annotations)
        return clone

    def build(self):
        return MethodAnnotationHandler(
            self._class_annotations + self._method_annotations
        )


class MethodAnnotationHandler(interfaces.AnnotationHandler):
    def __init__(self, method_annotations):
        self._method_annotations = list(method_annotations)

    @property
    def annotations(self):
        return iter(self._method_annotations)

    def handle_builder(self, request_builder):
        for annotation in self._method_annotations:
            annotation.modify_request(request_builder)


# TODO: Only decorate consumers
class MethodAnnotation(interfaces.Annotation):
    _http_method_blacklist = None
    _http_method_whitelist = None

    @staticmethod
    def _is_consumer_class(c):
        return utils.is_subclass(c, interfaces.Consumer)

    @classmethod
    def supports_http_method(cls, method):
        method = method.upper()
        if cls._http_method_blacklist is not None:
            return method not in cls._http_method_blacklist
        if cls._http_method_whitelist is not None:
            return method in cls._http_method_whitelist
        return True

    @classmethod
    def _is_relevant_for_builder(cls, builder):
        return cls.supports_http_method(builder[1].method)

    @classmethod
    def _is_static_call(cls, *args_, **kwargs):
        if super(MethodAnnotation, cls)._is_static_call(*args_, **kwargs):
            return True
        try:
            is_consumer_class = cls._is_consumer_class(args_[0])
        except IndexError:
            return False
        else:
            return is_consumer_class and not (kwargs or args_[1:])

    def _modify_request_definition(self, builder, kwargs):
        builder.method_handler_builder.add_annotation(self, **kwargs)

    def __call__(self, class_or_builder, **kwargs):
        if self._is_consumer_class(class_or_builder):
            builders = helpers.get_api_definitions(class_or_builder)
            builders = filter(self._is_relevant_for_builder, builders)

            for name, b in builders:
                self(b, is_class=True)
                helpers.set_api_definition(class_or_builder, name, b)
        elif isinstance(class_or_builder, interfaces.RequestDefinitionBuilder):
            self._modify_request_definition(class_or_builder, kwargs)
        return class_or_builder

    def modify_request(self, request_builder):
        pass


class _BaseRequestProperties(MethodAnnotation):
    _property_name = None
    _delimiter = None

    def __init__(self, arg, **kwargs):
        if isinstance(arg, list):
            self._values = dict(self._split(a) for a in arg)
        else:
            self._values = dict(arg, **kwargs)

    def _split(self, arg):
        return map(str.strip, arg.split(self._delimiter))

    def modify_request(self, request_builder):
        """Updates header contents."""
        request_builder.info[self._property_name].update(self._values)


# noinspection PyPep8Naming
class headers(_BaseRequestProperties):
    """
    A decorator that adds static headers for API calls.

    .. code-block:: python

        @headers({"User-Agent": "Uplink-Sample-App"})
        @get("/user")
        def get_user(self):
            \"""Get the current user\"""

    When used as a class decorator, :py:class:`headers` applies to
    all consumer methods bound to the class:

    .. code-block:: python

        @headers({"Accept": "application/vnd.github.v3.full+json"})
        class GitHub(Consumer):
            ...

    :py:class:`headers` takes the same arguments as :py:class:`dict`.

    Args:
        arg: A dict containing header values.
        **kwargs: More header values.
    """

    def __init__(self, arg=None, **kwargs):
        if isinstance(arg, str):
            key, value = self._split(arg)
            arg = {key: value}
        super(headers, self).__init__(arg or {}, **kwargs)

    @property
    def _delimiter(self):
        return ":"

    @property
    def _property_name(self):
        return "headers"


# noinspection PyPep8Naming
class params(_BaseRequestProperties):
    """
    A decorator that adds static query parameters for API calls.

    .. code-block:: python

        @params({"sort": "created"})
        @get("/user")
        def get_user(self):
            \"""Get the current user\"""

    When used as a class decorator, :py:class:`params` applies to
    all consumer methods bound to the class:

    .. code-block:: python

        @params({"client_id": "my-app-client-id"})
        class GitHub(Consumer):
            ...

    :py:class:`params` takes the same arguments as :py:class:`dict`.

    Args:
        arg: A dict containing query parameters.
        **kwargs: More query parameters.
    """

    def __init__(self, arg=None, **kwargs):
        if isinstance(arg, str):
            arg = arg.split("&")
        super(params, self).__init__(arg or {}, **kwargs)

    @property
    def _property_name(self):
        return "params"

    @property
    def _delimiter(self):
        return "="


# noinspection PyPep8Naming
class form_url_encoded(MethodAnnotation):
    """
    URL-encodes the request body.

    Used on POST/PUT/PATCH request. It url-encodes the body of the
    message and sets the appropriate ``Content-Type`` header. Further,
    each field argument should be annotated with
    :py:class:`uplink.Field`.

    Example:
        .. code-block:: python

            @form_url_encoded
            @post("/users/edit")
            def update_user(self, first_name: Field, last_name: Field):
                \"""Update the current user.\"""
    """

    _http_method_blacklist = {"GET"}
    _can_be_static = True

    # XXX: Let `requests` handle building urlencoded syntax.
    # def modify_request(self, request_builder):
    #     request_builder.info.headers(
    #         {"Content-Type": "application/x-www-form-urlencoded"}
    #     )


# noinspection PyPep8Naming
class multipart(MethodAnnotation):
    """
    Sends multipart form data.

    Multipart requests are commonly used to upload files to a server.
    Further, annotate each part argument with :py:class:`Part`.

    Example:

        .. code-block:: python

            @multipart
            @put(/user/photo")
            def update_user(self, photo: Part, description: Part):
                \"""Upload a user profile photo.\"""
    """

    _http_method_blacklist = {"GET"}
    _can_be_static = True

    # XXX: Let `requests` handle building multipart syntax.
    # def modify_request(self, request_builder):
    #     request_builder.info.headers(
    #         {"Content-Type": "multipart/form-data"}
    #     )


# noinspection PyPep8Naming
class json(MethodAnnotation):
    """Use as a decorator to make JSON requests.

    You can annotate a method argument with :py:class:`uplink.Body`,
    which indicates that the argument's value should become the
    request's body. :py:class:`uplink.Body` has to be either a dict or a
    subclass of py:class:`abc.Mapping`.

    Example:
        .. code-block:: python

            @json
            @patch(/user")
            def update_user(self, **info: Body):
                \"""Update the current user.\"""

    You can alternatively use the :py:class:`uplink.Field` annotation to
    specify JSON fields separately, across multiple arguments:

    Example:
        .. code-block:: python

            @json
            @patch(/user")
            def update_user(self, name: Field, email: Field("e-mail")):
                \"""Update the current user.\"""

    Further, to set a nested field, you can specify the path of the
    target field with a tuple of strings as the first argument of
    :py:class:`uplink.Field`.

    Example:
        Consider a consumer method that sends a PATCH request with a JSON
        body of the following format:

        .. code-block:: json
            :emphasize-lines: 3

            {
                user: {
                    name: "<User's Name>"
                },
            }

        The tuple :py:obj:`("user", "name")` specifies the path to the
        highlighted inner field:

        .. code-block:: python
            :emphasize-lines: 5

            @json
            @patch(/user")
            def update_user(
                            self,
                            new_name: Field(("user", "name"))
            ):
                \"""Update the current user.\"""
    """

    _http_method_blacklist = {"GET"}
    _can_be_static = True

    @staticmethod
    def _sequence_path_resolver(path, value, body):
        if not path:
            raise ValueError("Path sequence cannot be empty.")
        for name in path[:-1]:
            body = body.setdefault(name, {})
            if not isinstance(body, abc.Mapping):
                raise ValueError(
                    "Failed to set nested JSON attribute '%s': "
                    "parent field '%s' is not a JSON object." % (path, name)
                )
        body[path[-1]] = value

    def modify_request(self, request_builder):
        """Modifies JSON request."""
        request_builder.add_transaction_hook(self._hook)

    @classmethod
    def set_json_body(cls, request_builder):
        old_body = request_builder.info.pop("data", {})
        if isinstance(old_body, abc.Mapping):
            body = request_builder.info.setdefault("json", {})
            for path in old_body:
                if isinstance(path, tuple):
                    cls._sequence_path_resolver(path, old_body[path], body)
                else:
                    body[path] = old_body[path]
        else:
            request_builder.info.setdefault("json", old_body)

    __hook = None

    @property
    def _hook(self):
        if self.__hook is None:
            self.__hook = hooks.RequestAuditor(self.set_json_body)
        return self.__hook


# noinspection PyPep8Naming
class timeout(MethodAnnotation):
    """
    Time to wait for a server response before giving up.

    When used on other decorators it specifies how long (in secs) a
    decorator should wait before giving up.

    Example:
        .. code-block:: python

            @timeout(60)
            @get("/user/posts")
            def get_posts(self):
                \"""Fetch all posts for the current users.\"""

    When used as a class decorator, :py:class:`timeout` applies to all
    consumer methods bound to the class.

    Args:
        seconds (int): An integer used to indicate how long should the
            request wait.
    """

    def __init__(self, seconds):
        self._seconds = seconds

    def modify_request(self, request_builder):
        """Modifies request timeout."""
        request_builder.info["timeout"] = self._seconds


# noinspection PyPep8Naming
class args(MethodAnnotation):
    """
    Annotate method arguments for Python 2.7 compatibility.

    Arrange annotations in the same order as their corresponding
    function arguments.

    Example:
        .. code-block:: python

            @args(Path, Query)
            @get("/users/{username})
            def get_user(self, username, visibility):
                \"""Get a specific user.\"""

    Use keyword args to target specific method parameters.

    Example:
        .. code-block:: python

            @args(visibility=Query)
            @get("/users/{username})
            def get_user(self, username, visibility):
                \"""Get a specific user.\"""

    Args:
        *annotations: Any number of annotations.
        **more_annotations: More annotations, targeting specific method
           arguments.
    """

    def __init__(self, *annotations, **more_annotations):
        self._annotations = annotations
        self._more_annotations = more_annotations

    def __call__(self, obj):
        if inspect.isfunction(obj):
            handler = arguments.ArgumentAnnotationHandlerBuilder.from_func(obj)
            self._helper(handler)
            return obj
        else:
            return super(args, self).__call__(obj)

    def _helper(self, builder):
        builder.set_annotations(self._annotations, **self._more_annotations)

    def modify_request_definition(self, request_definition_builder):
        """Modifies dynamic requests with given annotations"""
        self._helper(request_definition_builder.argument_handler_builder)


class _InjectableMethodAnnotation(MethodAnnotation):
    def modify_request(self, request_builder):
        request_builder.add_transaction_hook(self)


class _BaseHandlerAnnotation(_InjectableMethodAnnotation):
    def __new__(cls, func=None, *args, **kwargs):
        if func is None:
            return lambda f: cls(f, *args, **kwargs)
        self = super(_BaseHandlerAnnotation, cls).__new__(cls)
        functools.update_wrapper(self, func)
        return self


# noinspection PyPep8Naming
class response_handler(_BaseHandlerAnnotation, hooks.ResponseHandler):
    """
    A decorator for creating custom response handlers.

    To register a function as a custom response handler, decorate the
    function with this class. The decorated function should accept a single
    positional argument, an HTTP response object:

    Example:
        .. code-block:: python

            @response_handler
            def raise_for_status(response):
                response.raise_for_status()
                return response

    Then, to apply custom response handling to a request method, simply
    decorate the method with the registered response handler:

    Example:
        .. code-block:: python

            @raise_for_status
            @get("/user/posts")
            def get_posts(self):
                \"""Fetch all posts for the current users.\"""

    To apply custom response handling on all request methods of a
    :py:class:`uplink.Consumer` subclass, simply decorate the class with
    the registered response handler:

    Example:
        .. code-block:: python

            @raise_for_status
            class GitHub(Consumer):
               ...

    Lastly, the decorator supports the optional argument
    :obj:`requires_consumer`. When this option is set to :obj:`True`,
    the registered callback should accept a reference to the
    :class:`~Consumer` instance as its leading argument:

    Example:
        .. code-block:: python

            @response_handler(requires_consumer=True)
            def raise_for_status(consumer, response):
                ...

    .. versionadded:: 0.4.0
    """


# noinspection PyPep8Naming
class error_handler(_BaseHandlerAnnotation, hooks.ExceptionHandler):
    """
    A decorator for creating custom error handlers.

    To register a function as a custom error handler, decorate the
    function with this class. The decorated function should accept three
    positional arguments: (1) the type of the exception, (2) the
    exception instance raised, and (3) a traceback instance.

    Example:
        .. code-block:: python

            @error_handler
            def raise_api_error(exc_type, exc_val, exc_tb):
                # wrap client error with custom API error
                ...

    Then, to apply custom error handling to a request method, simply
    decorate the method with the registered error handler:

    Example:
        .. code-block:: python

            @raise_api_error
            @get("/user/posts")
            def get_posts(self):
                \"""Fetch all posts for the current users.\"""

    To apply custom error handling on all request methods of a
    :py:class:`uplink.Consumer` subclass, simply decorate the class with
    the registered error handler:

    Example:
        .. code-block:: python

            @raise_api_error
            class GitHub(Consumer):
               ...

    Lastly, the decorator supports the optional argument
    :obj:`requires_consumer`. When this option is set to :obj:`True`,
    the registered callback should accept a reference to the
    :class:`~Consumer` instance as its leading argument:

    Example:
        .. code-block:: python

             @error_handler(requires_consumer=True)
             def raise_api_error(consumer, exc_type, exc_val, exc_tb):
                ...

    .. versionadded:: 0.4.0

    Note:
        Error handlers can not completely suppress exceptions. The
        original exception is thrown if the error handler doesn't throw
        anything.
    """


# noinspection PyPep8Naming
class inject(_InjectableMethodAnnotation, hooks.TransactionHookChain):
    """
    A decorator that applies one or more hooks to a request method.

    .. versionadded:: 0.4.0
    """



import pytest

# Local imports
from uplink import hooks, arguments
from uplink.converters import keys


inject_args = pytest.mark.parametrize("args", (["arg1", "arg2", "arg3"],))


@pytest.fixture
def argument_mock(mocker):
    return mocker.Mock(spec=arguments.ArgumentAnnotation)


@pytest.fixture
def named_argument_mock(mocker):
    return mocker.Mock(spec=arguments.NamedArgument)


class ArgumentTestCase(object):
    @property
    def type_cls(self):
        raise NotImplementedError

    @property
    def expected_converter_key(self):
        raise NotImplementedError

    @property
    def expected_can_be_static(self):
        return True

    def test_converter_type(self):
        assert self.type_cls().converter_key == self.expected_converter_key

    def test_is_static(self):
        assert self.type_cls._can_be_static == self.expected_can_be_static

    def test_static_call(self, mocker, request_definition_builder):
        request_definition_builder = self.type_cls(request_definition_builder)
        builder = request_definition_builder.argument_handler_builder
        builder.add_annotation.assert_called_with(mocker.ANY)


class FuncDecoratorTestCase(object):
    def test_static_call_with_function(self):
        def func(a1, a2):
            return a1, a2

        output = self.type_cls(func)
        assert output is func

    def test_equals(self):
        assert isinstance(
            self.type_cls().with_value("hello"), hooks.TransactionHook
        )


class TestArgumentAnnotationHandlerBuilder(object):
    def test_from_func(self):
        def func(_):
            pass

        handler = arguments.ArgumentAnnotationHandlerBuilder.from_func(func)
        another_handler = arguments.ArgumentAnnotationHandlerBuilder.from_func(
            func
        )
        assert handler is another_handler

    @inject_args
    def test_missing_arguments(self, args):
        builder = arguments.ArgumentAnnotationHandlerBuilder(None, args, False)
        assert list(builder.missing_arguments) == args

    @inject_args
    def test_remaining_args_count(self, args):
        builder = arguments.ArgumentAnnotationHandlerBuilder(None, args, False)
        assert builder.remaining_args_count == len(args)

    @inject_args
    def test_add_annotation_without_name(self, mocker, argument_mock, args):
        builder = arguments.ArgumentAnnotationHandlerBuilder(None, args, False)
        builder.listener = mocker.stub()
        builder.add_annotation(argument_mock)
        builder.listener.assert_called_with(argument_mock)
        assert args[0] not in builder.missing_arguments

    @inject_args
    def test_add_annotation_with_name(self, mocker, argument_mock, args):
        builder = arguments.ArgumentAnnotationHandlerBuilder(None, args, False)
        builder.listener = mocker.stub()
        builder.add_annotation(argument_mock, name=args[-1])
        builder.listener.assert_called_with(argument_mock)
        assert args[-1] not in builder.missing_arguments

    @inject_args
    def test_add_named_annotation_without_name(
        self, mocker, named_argument_mock, args
    ):
        builder = arguments.ArgumentAnnotationHandlerBuilder(None, args, False)
        named_argument_mock.name = None
        builder.listener = mocker.stub()
        builder.add_annotation(named_argument_mock)
        builder.listener.assert_called_with(named_argument_mock)
        assert named_argument_mock.name == args[0]
        assert args[0] not in builder.missing_arguments

    @inject_args
    def test_add_annotation_class(self, mocker, args):
        builder = arguments.ArgumentAnnotationHandlerBuilder(None, args, False)
        builder.listener = mocker.stub()
        argument = builder.add_annotation(arguments.ArgumentAnnotation())
        builder.listener.assert_called_with(argument)
        assert args[0] not in builder.missing_arguments

    @inject_args
    def test_add_annotation_with_name_not_recognized(self, argument_mock, args):
        def dummy():
            pass

        assert -1 not in args
        builder = arguments.ArgumentAnnotationHandlerBuilder(dummy, args, False)
        with pytest.raises(arguments.ArgumentNotFound):
            builder.add_annotation(argument_mock, name=-1)

    def test_add_annotation_with_no_missing_arguments(self, argument_mock):
        def dummy():
            pass

        builder = arguments.ArgumentAnnotationHandlerBuilder(dummy, [], False)
        with pytest.raises(arguments.ExhaustedArguments):
            builder.add_annotation(argument_mock)

    def test_add_annotation_that_is_not_an_annotation(self):
        def dummy():
            pass

        builder = arguments.ArgumentAnnotationHandlerBuilder(
            dummy, ["arg1"], False
        )
        builder.add_annotation(type, "arg1")
        assert builder.remaining_args_count == 1

    @inject_args
    def test_set_annotations(self, mocker, argument_mock, args):
        builder = arguments.ArgumentAnnotationHandlerBuilder(None, args, False)
        builder.listener = mocker.stub()
        builder.set_annotations((argument_mock,))
        builder.listener.assert_called_with(argument_mock)
        assert args[0] not in builder.missing_arguments

    @inject_args
    def test_set_annotations_with_dict(self, mocker, argument_mock, args):
        builder = arguments.ArgumentAnnotationHandlerBuilder(None, args, False)
        builder.listener = mocker.stub()
        builder.set_annotations(**{args[0]: argument_mock})
        builder.listener.assert_called_with(argument_mock)
        assert args[0] not in builder.missing_arguments

    def test_is_done(self, argument_mock):
        builder = arguments.ArgumentAnnotationHandlerBuilder(
            None, ("arg1",), False
        )
        assert not builder.is_done()
        builder.add_annotation(argument_mock)
        assert builder.is_done()


class TestArgumentAnnotationHandler(object):
    def test_get_relevant_arguments(self):
        args = {"arg1": "value1"}
        annotation_handler = arguments.ArgumentAnnotationHandler(None, args)
        relevant = annotation_handler.get_relevant_arguments(args)
        assert list(relevant) == list(args.items())

    def test_handle_call(self, request_builder, mocker):
        def dummy(arg1):
            return arg1

        request_builder.get_converter.return_value = dummy
        get_call_args = mocker.patch("uplink.utils.get_call_args")
        get_call_args.return_value = {"arg1": "hello"}
        annotation = mocker.Mock(arguments.ArgumentAnnotation)
        handlers = arguments.ArgumentAnnotationHandler(
            dummy, {"arg1": annotation}
        )
        handlers.handle_call(request_builder, (), {})
        annotation.modify_request.assert_called_with(request_builder, "hello")

    @inject_args
    def test_annotations(self, args):
        annotations = ["annotation"] * len(args)
        arg_dict = dict(zip(args, annotations))
        annotation_handler = arguments.ArgumentAnnotationHandler(None, arg_dict)
        assert list(annotation_handler.annotations) == annotations


class TestArgumentAnnotation(object):
    def test_call(self, request_definition_builder):
        annotation = arguments.ArgumentAnnotation()
        return_value = annotation(request_definition_builder)
        handler_builder = request_definition_builder.argument_handler_builder
        handler_builder.add_annotation.assert_called_with(annotation)
        assert return_value is request_definition_builder


class TestTypedArgument(object):
    def test_type(self):
        assert arguments.TypedArgument("hello").type == "hello"

    def test_set_type(self):
        annotation = arguments.TypedArgument()
        assert annotation.type is None
        annotation.type = "type"
        assert annotation.type == "type"

    def test_set_type_with_type_already_set(self):
        annotation = arguments.TypedArgument("type")
        with pytest.raises(AttributeError):
            annotation.type = "new type"


class TestNamedArgument(object):
    def test_name(self):
        assert arguments.NamedArgument("name").name == "name"

    def test_set_name(self):
        annotation = arguments.NamedArgument()
        assert annotation.name is None
        annotation.name = "name"
        assert annotation.name == "name"

    def test_set_name_with_name_already_set(self):
        annotation = arguments.NamedArgument("name")
        with pytest.raises(AttributeError):
            annotation.name = "new name"


class TestPath(ArgumentTestCase):
    type_cls = arguments.Path
    expected_converter_key = keys.CONVERT_TO_STRING

    def test_modify_request_definition(self, request_definition_builder):
        arguments.Path("name").modify_request_definition(
            request_definition_builder
        )
        request_definition_builder.uri.add_variable.assert_called_with("name")

    def test_modify_request(self, request_builder):
        arguments.Path("name").modify_request(request_builder, "value")
        request_builder.set_url_variable.assert_called_with({"name": "value"})


class TestQuery(ArgumentTestCase, FuncDecoratorTestCase):
    type_cls = arguments.Query
    expected_converter_key = keys.Sequence(keys.CONVERT_TO_STRING)

    def test_modify_request(self, request_builder):
        arguments.Query("name").modify_request(request_builder, "value")
        assert request_builder.info["params"] == {"name": "value"}

    def test_modify_request_with_encoded(self, request_builder):
        arguments.Query("name", encoded=True).modify_request(
            request_builder, "value"
        )
        assert request_builder.info["params"] == "name=value"

    def test_modify_request_with_mismatched_encoding(self, request_builder):
        arguments.Query("name", encoded=True).modify_request(
            request_builder, "value"
        )
        with pytest.raises(arguments.Query.QueryStringEncodingError):
            arguments.Query("name2", encoded=False).modify_request(
                request_builder, "value2"
            )

    def test_skip_none(self, request_builder):
        arguments.Query("name").modify_request(request_builder, None)
        assert request_builder.info["params"] == {}

    def test_encode_none(self, request_builder):
        arguments.Query("name", encode_none="null").modify_request(
            request_builder, None
        )
        assert request_builder.info["params"] == {"name": "null"}

    def test_converter_key_with_encoded(self):
        query = arguments.Query("name", encoded=True)
        assert query.converter_key == keys.CONVERT_TO_STRING

    def test_converter_type(self):
        query = arguments.Query("name", encoded=False)
        assert query.converter_key == keys.Sequence(keys.CONVERT_TO_STRING)


class TestQueryMap(ArgumentTestCase, FuncDecoratorTestCase):
    type_cls = arguments.QueryMap
    expected_converter_key = keys.Map(TestQuery.expected_converter_key)

    def test_modify_request(self, request_builder):
        arguments.QueryMap().modify_request(request_builder, {"hello": "world"})
        assert request_builder.info["params"] == {"hello": "world"}

    def test_modify_request_with_encoded(self, request_builder):
        arguments.QueryMap(encoded=True).modify_request(
            request_builder, {"name": "value"}
        )
        assert request_builder.info["params"] == "name=value"

    def test_converter_key_with_encoded(self):
        query = arguments.QueryMap(encoded=True)
        assert query.converter_key == keys.Map(keys.CONVERT_TO_STRING)

    def test_converter_type(self):
        query = arguments.QueryMap(encoded=False)
        assert query.converter_key == keys.Map(
            keys.Sequence(keys.CONVERT_TO_STRING)
        )


class TestHeader(ArgumentTestCase, FuncDecoratorTestCase):
    type_cls = arguments.Header
    expected_converter_key = keys.CONVERT_TO_STRING

    def test_modify_request(self, request_builder):
        arguments.Header("hello").modify_request(request_builder, "world")
        assert request_builder.info["headers"] == {"hello": "world"}

    def test_skip_none(self, request_builder):
        arguments.Header("hello").modify_request(request_builder, None)
        assert request_builder.info["headers"] == {}


class TestHeaderMap(ArgumentTestCase, FuncDecoratorTestCase):
    type_cls = arguments.HeaderMap
    expected_converter_key = keys.Map(TestHeader.expected_converter_key)

    def test_modify_request(self, request_builder):
        arguments.HeaderMap().modify_request(
            request_builder, {"hello": "world"}
        )
        assert request_builder.info["headers"] == {"hello": "world"}


class TestField(ArgumentTestCase):
    type_cls = arguments.Field
    expected_converter_key = keys.CONVERT_TO_REQUEST_BODY

    def test_modify_request(self, request_builder):
        arguments.Field("hello").modify_request(request_builder, "world")
        assert request_builder.info["data"]["hello"] == "world"

    def test_modify_request_failure(self, request_builder):
        request_builder.info["data"] = object()
        with pytest.raises(arguments.Field.FieldAssignmentFailed):
            arguments.Field("hello").modify_request(request_builder, "world")


class TestFieldMap(ArgumentTestCase):
    type_cls = arguments.FieldMap
    expected_converter_key = keys.Map(TestField.expected_converter_key)

    def test_modify_request(self, request_builder):
        arguments.FieldMap().modify_request(request_builder, {"hello": "world"})
        assert request_builder.info["data"] == {"hello": "world"}

    def test_modify_request_failure(self, request_builder):
        request_builder.info["data"] = object()
        with pytest.raises(arguments.FieldMap.FieldMapUpdateFailed):
            arguments.FieldMap().modify_request(request_builder, {})


class TestPart(ArgumentTestCase):
    type_cls = arguments.Part
    expected_converter_key = keys.CONVERT_TO_REQUEST_BODY

    def test_modify_request(self, request_builder):
        arguments.Part("hello").modify_request(request_builder, "world")
        assert request_builder.info["files"] == {"hello": "world"}


class TestPartMap(ArgumentTestCase):
    type_cls = arguments.PartMap
    expected_converter_key = keys.Map(TestPart.expected_converter_key)

    def test_modify_request(self, request_builder):
        arguments.PartMap().modify_request(request_builder, {"hello": "world"})
        assert request_builder.info["files"] == {"hello": "world"}


class TestBody(ArgumentTestCase):
    type_cls = arguments.Body
    expected_converter_key = keys.CONVERT_TO_REQUEST_BODY

    def test_modify_request(self, request_builder):
        # Verify with dict
        arguments.Body().modify_request(request_builder, {"hello": "world"})
        assert request_builder.info["data"] == {"hello": "world"}

        # Verify with non-mapping
        body = object()
        arguments.Body().modify_request(request_builder, body)
        assert request_builder.info["data"] is body


class TestUrl(ArgumentTestCase):
    type_cls = arguments.Url
    expected_converter_key = keys.CONVERT_TO_STRING

    def test_modify_request_definition(self, request_definition_builder):
        arguments.Url().modify_request_definition(request_definition_builder)
        assert request_definition_builder.uri.is_dynamic

    def test_modify_request_definition_failure(
        self, mocker, request_definition_builder
    ):
        is_dynamic_mock = mocker.PropertyMock(side_effect=ValueError())
        type(request_definition_builder.uri).is_dynamic = is_dynamic_mock
        request_definition_builder.__name__ = "dummy"
        with pytest.raises(arguments.Url.DynamicUrlAssignmentFailed):
            arguments.Url().modify_request_definition(
                request_definition_builder
            )

    def test_modify_request(self, request_builder):
        arguments.Url().modify_request(request_builder, "/some/path")
        assert request_builder.relative_url == "/some/path"


class TestTimeout(ArgumentTestCase, FuncDecoratorTestCase):
    type_cls = arguments.Timeout
    expected_converter_key = keys.Identity()

    def test_modify_request(self, request_builder):
        arguments.Timeout().modify_request(request_builder, 10)
        assert request_builder.info["timeout"] == 10


class TestContext(ArgumentTestCase, FuncDecoratorTestCase):
    type_cls = arguments.Context
    expected_converter_key = keys.Identity()

    def test_modify_request(self, request_builder):
        arguments.Context("key").modify_request(request_builder, "value")
        assert request_builder.context["key"] == "value"


class TestContextMap(ArgumentTestCase, FuncDecoratorTestCase):
    type_cls = arguments.ContextMap
    expected_converter_key = keys.Identity()

    def test_modify_request(self, request_builder):
        arguments.ContextMap().modify_request(request_builder, {"key": "value"})
        assert request_builder.context == {"key": "value"}

    def test_modify_request_not_mapping(self, request_builder):
        with pytest.raises(TypeError):
            arguments.ContextMap().modify_request(request_builder, "value")



# Local imports
from uplink import Consumer, PartMap, post, multipart

# Constants
BASE_URL = "https://example.com/"


def test_without_converter(mock_response, mock_client):
    class Calendar(Consumer):
        @multipart
        @post("/attachments", args={"files": PartMap})
        def upload_attachments(self, **files):
            pass

    mock_client.with_response(mock_response)
    calendar = Calendar(base_url=BASE_URL, client=mock_client)
    file = object()

    # Run
    calendar.upload_attachments(file=file)

    # Assertion: should not convert if converter is None
    request = mock_client.history[0]
    assert request.files == {"file": file}



class AnnotationMeta(type):
    def __call__(cls, *args, **kwargs):
        if cls._can_be_static and cls._is_static_call(*args, **kwargs):
            self = super(AnnotationMeta, cls).__call__()
            self(args[0])
            return args[0]
        else:
            return super(AnnotationMeta, cls).__call__(*args, **kwargs)


class _Annotation(object):
    _can_be_static = False

    def modify_request_definition(self, request_definition_builder):
        pass

    @classmethod
    def _is_static_call(cls, *args, **kwargs):
        try:
            is_builder = isinstance(args[0], RequestDefinitionBuilder)
        except IndexError:
            return False
        else:
            return is_builder and not (kwargs or args[1:])


Annotation = AnnotationMeta("Annotation", (_Annotation,), {})


class AnnotationHandlerBuilder(object):
    __listener = None

    @property
    def listener(self):
        return self.__listener

    @listener.setter
    def listener(self, listener):
        self.__listener = listener

    def add_annotation(self, annotation, *args, **kwargs):
        if self.__listener is not None:
            self.__listener(annotation)

    def is_done(self):
        return True

    def build(self):
        raise NotImplementedError


class AnnotationHandler(object):
    @property
    def annotations(self):
        raise NotImplementedError


class UriDefinitionBuilder(object):
    @property
    def is_static(self):
        raise NotImplementedError

    @property
    def is_dynamic(self):
        raise NotImplementedError

    @is_dynamic.setter
    def is_dynamic(self, is_dynamic):
        raise NotImplementedError

    def add_variable(self, name):
        raise NotImplementedError

    @property
    def remaining_variables(self):
        raise NotImplementedError

    def build(self):
        raise NotImplementedError


class RequestDefinitionBuilder(object):
    @property
    def method(self):
        raise NotImplementedError

    @property
    def uri(self):
        raise NotImplementedError

    @property
    def argument_handler_builder(self):
        raise NotImplementedError

    @property
    def method_handler_builder(self):
        raise NotImplementedError

    def update_wrapper(self, wrapper):
        raise NotImplementedError

    def build(self):
        raise NotImplementedError

    def copy(self):
        raise NotImplementedError


class RequestDefinition(object):
    def make_converter_registry(self, converters):
        raise NotImplementedError

    def define_request(self, request_builder, func_args, func_kwargs):
        raise NotImplementedError


class CallBuilder(object):
    @property
    def client(self):
        raise NotImplementedError

    @property
    def base_url(self):
        raise NotImplementedError

    @property
    def converters(self):
        raise NotImplementedError

    @property
    def hooks(self):
        raise NotImplementedError

    def add_hook(self, hook, *more_hooks):
        raise NotImplementedError

    @property
    def auth(self):
        raise NotImplementedError

    def build(self, definition):
        raise NotImplementedError


class Auth(object):
    def __call__(self, request_builder):
        raise NotImplementedError


class Consumer(object):
    @property
    def session(self):
        raise NotImplementedError



# Local imports
from uplink.__about__ import __version__
from uplink._extras import install, load_entry_points as _load_entry_points
from uplink import returns, types
from uplink.clients import AiohttpClient, RequestsClient, TwistedClient

# todo: remove this in v1.0.0
from uplink.converters import MarshmallowConverter

from uplink.builder import build, Consumer
from uplink.commands import get, head, put, post, patch, delete
from uplink.exceptions import (
    Error,
    UplinkBuilderError,
    InvalidRequestDefinition,
    AnnotationError,
)
from uplink.decorators import (
    headers,
    params,
    form_url_encoded,
    multipart,
    json,
    timeout,
    args,
    response_handler,
    error_handler,
    inject,
)
from uplink.models import loads, dumps
from uplink.arguments import (
    Path,
    Query,
    QueryMap,
    Header,
    HeaderMap,
    Field,
    FieldMap,
    Part,
    PartMap,
    Body,
    Url,
    Timeout,
    Context,
)
from uplink.ratelimit import ratelimit
from uplink.retry import retry

__all__ = [
    "__version__",
    "install",
    "returns",
    "types",
    "AiohttpClient",
    "RequestsClient",
    "TwistedClient",
    "MarshmallowConverter",
    "build",
    "Consumer",
    "get",
    "head",
    "put",
    "post",
    "patch",
    "delete",
    "Error",
    "UplinkBuilderError",
    "InvalidRequestDefinition",
    "AnnotationError",
    "headers",
    "params",
    "form_url_encoded",
    "multipart",
    "json",
    "timeout",
    "args",
    "response_handler",
    "error_handler",
    "inject",
    "loads",
    "dumps",
    "Path",
    "Query",
    "QueryMap",
    "Header",
    "HeaderMap",
    "Field",
    "FieldMap",
    "Part",
    "PartMap",
    "Body",
    "Url",
    "Timeout",
    "Context",
    "retry",
    "ratelimit",
]

_load_entry_points()



# Standard library imports
import contextlib

# Third-party imports
import aiohttp
import asyncio
import pytest

# Local imports
from uplink.clients import (
    AiohttpClient,
    aiohttp_,
    register,
    io,
)


@contextlib.contextmanager
def _patch(obj, attr, value):
    if obj is not None:
        old_value = getattr(obj, attr)
        setattr(obj, attr, value)
    yield
    if obj is not None:
        setattr(obj, attr, old_value)


@pytest.fixture
def aiohttp_session_mock(mocker):
    return mocker.Mock(spec=aiohttp.ClientSession)


class AsyncMock(object):
    def __init__(self, result=None):
        self._result = result
        self._calls = 0

    async def __call__(self, *args, **kwargs):
        self._calls += 1
        f = asyncio.Future()
        f.set_result(self._result)
        return f

    @property
    def called(self):
        return self._calls > 0


class TestAiohttp(object):
    def test_init_when_aiohttp_is_not_installed(self):
        with _patch(aiohttp_, "aiohttp", None):
            with pytest.raises(NotImplementedError):
                AiohttpClient()

    def test_init_with_session_None(self, mocker):
        mocker.spy(AiohttpClient, "_create_session")
        AiohttpClient(kwarg1="value")
        AiohttpClient._create_session.assert_called_with(kwarg1="value")

    def test_get_client(self, aiohttp_session_mock):
        client = register.get_client(aiohttp_session_mock)
        assert isinstance(client, aiohttp_.AiohttpClient)

    @pytest.mark.asyncio
    async def test_request_send(self, mocker, aiohttp_session_mock):
        # Setup
        expected_response = mocker.Mock()

        async def request(*args, **kwargs):
            return expected_response

        aiohttp_session_mock.request = request
        client = aiohttp_.AiohttpClient(aiohttp_session_mock)

        # Run
        response = await client.send((1, 2, {}))

        # Verify
        assert response == expected_response

    @pytest.mark.asyncio
    async def test_callback(self, mocker, aiohttp_session_mock):
        # Setup
        expected_response = mocker.Mock(spec=aiohttp_.aiohttp.ClientResponse)
        expected_response.text = AsyncMock()

        async def request(*args, **kwargs):
            return expected_response

        aiohttp_session_mock.request = request
        client = aiohttp_.AiohttpClient(aiohttp_session_mock)

        # Run
        async def call():
            response = await client.send((1, 2, {}))
            response = await client.apply_callback(lambda x: 2, response)
            return response

        value = await call()

        # Verify
        assert value == 2
        assert expected_response.text.called

    def test_wrap_callback(self, mocker):
        # Setup
        c = AiohttpClient()
        mocker.spy(c, "_sync_callback_adapter")

        # Run: with callback that is not a coroutine
        def callback(*_):
            pass

        c.wrap_callback(callback)

        # Verify: Should wrap it
        c._sync_callback_adapter.assert_called_with(callback)

        # Run: with coroutine callback
        async def awaitable_callback():
            pass

        assert c.wrap_callback(awaitable_callback) is awaitable_callback

    @pytest.mark.asyncio
    async def test_threaded_callback(self, mocker):
        def callback(response):
            return response

        # Mock response.
        response = mocker.Mock(spec=aiohttp_.aiohttp.ClientResponse)
        response.text = AsyncMock()

        # Run
        new_callback = aiohttp_.threaded_callback(callback)
        return_value = await new_callback(response)

        # Verify
        assert response.text.called
        assert return_value == response

        # Run: Verify with callback that returns new value
        def callback(*_):
            return 1

        new_callback = aiohttp_.threaded_callback(callback)
        value = await new_callback(response)
        assert value == 1
        assert response.text.called

        # Run: Verify with response that is not ClientResponse (should not be wrapped)
        response = mocker.Mock()
        await new_callback(response)
        assert not response.text.called

    def test_threaded_coroutine(self):
        async def coroutine():
            return 1

        threaded_coroutine = aiohttp_.ThreadedCoroutine(coroutine)

        # Run -- should block
        response = threaded_coroutine()

        # Verify
        assert response == 1

    def test_threaded_response(self, mocker):
        async def coroutine():
            return 1

        def not_a_coroutine():
            return 2

        response = mocker.Mock()
        response.coroutine = coroutine
        response.not_coroutine = not_a_coroutine
        threaded_response = aiohttp_.ThreadedResponse(response)

        # Run
        threaded_coroutine = threaded_response.coroutine
        return_value = threaded_coroutine()

        # Verify
        assert isinstance(threaded_coroutine, aiohttp_.ThreadedCoroutine)
        assert return_value == 1
        assert threaded_response.not_coroutine is not_a_coroutine

    @pytest.mark.asyncio
    async def test_create(self, mocker):
        session_cls_mock = mocker.patch("aiohttp.ClientSession")
        positionals = [1]
        keywords = {"keyword": 2}

        # Run: Create client
        client = aiohttp_.AiohttpClient.create(*positionals, **keywords)

        # Verify: session hasn't been created yet.
        assert not session_cls_mock.called

        # Run: Get session
        await client.session()

        # Verify: session created with args
        session_cls_mock.assert_called_with(*positionals, **keywords)

    @pytest.mark.asyncio
    async def test_close_auto_created_session(self, mocker):
        # Setup
        import gc

        mock_session = mocker.Mock(spec=aiohttp.ClientSession)
        session_cls_mock = mocker.patch("aiohttp.ClientSession")
        session_cls_mock.return_value = mock_session

        positionals = [1]
        keywords = {"keyword": 2}

        # Run: Create client
        client = aiohttp_.AiohttpClient.create(*positionals, **keywords)

        # Run: Get session
        await client.session()

        # Verify: session created with args
        session_cls_mock.assert_called_with(*positionals, **keywords)

        # Verify: session closed on garbage collection
        del client
        gc.collect()
        session_cls_mock.return_value.close.assert_called_with()

    def test_exceptions(self):
        import aiohttp

        exceptions = aiohttp_.AiohttpClient.exceptions

        with pytest.raises(exceptions.BaseClientException):
            raise aiohttp.ClientError()

        with pytest.raises(exceptions.BaseClientException):
            # Test polymorphism
            raise aiohttp.InvalidURL("invalid")

        with pytest.raises(exceptions.ConnectionError):
            raise aiohttp.ClientConnectionError()

        with pytest.raises(exceptions.ConnectionTimeout):
            raise aiohttp.ClientConnectorError.__new__(
                aiohttp.ClientConnectorError
            )

        with pytest.raises(exceptions.ServerTimeout):
            raise aiohttp.ServerTimeoutError()

        with pytest.raises(exceptions.SSLError):
            raise aiohttp.ClientSSLError.__new__(aiohttp.ClientSSLError)

        with pytest.raises(exceptions.InvalidURL):
            raise aiohttp.InvalidURL("invalid")

    def test_io(self):
        assert isinstance(aiohttp_.AiohttpClient.io(), io.AsyncioStrategy)



class _UnmappedClientException(BaseException):
    pass


class Exceptions(object):
    """Enum of standard HTTP client exceptions."""

    BaseClientException = _UnmappedClientException
    """Base class for client errors."""

    ConnectionError = _UnmappedClientException
    """A connection error occurred."""

    ConnectionTimeout = _UnmappedClientException
    """The request timed out while trying to connect to the remote server.

    Requests that produce this error are typically safe to retry.
    """

    ServerTimeout = _UnmappedClientException
    """The server did not send any data in the allotted amount of time."""

    SSLError = _UnmappedClientException
    """An SSL error occurred."""

    InvalidURL = _UnmappedClientException
    """The URL provided was somehow invalid."""



class Error(Exception):
    """Base exception for this package"""

    message = None

    def __str__(self):
        return str(self.message)


class UplinkBuilderError(Error):
    """Something went wrong while building a service."""

    message = "`%s`: %s"

    def __init__(self, class_name, definition_name, error):
        fullname = class_name + "." + definition_name
        self.message = self.message % (fullname, error)
        self.error = error


class InvalidRequestDefinition(Error):
    """Something went wrong when building the request definition."""


class AnnotationError(Error):
    """Something went wrong with an annotation."""



# Third-party imports
import pytest
import pytest_twisted

# Local imports.
from uplink import get, Consumer, retry
from uplink.clients import io

# Constants
BASE_URL = "https://api.github.com/"


def backoff_once():
    yield 0.1


backoff_default = retry.backoff.exponential(multiplier=0.1, minimum=0.1)


class GitHub(Consumer):
    @retry(max_attempts=2, backoff=backoff_default)
    @get("/users/{user}")
    def get_user(self, user):
        pass

    @retry(max_attempts=3, backoff=backoff_once)
    @get("repos/{user}/{repo}/issues/{issue}")
    def get_issue(self, user, repo, issue):
        pass

    @retry(
        stop=retry.stop.after_attempt(3), on_exception=retry.CONNECTION_TIMEOUT
    )
    @get("repos/{user}/{repo}/project/{project}")
    def get_project(self, user, repo, project):
        pass

    @retry(when=retry.when.status_5xx(), backoff=backoff_default)
    @get("repos/{user}/{repo}/issues")
    def get_issues(self, user, repo):
        pass


# Tests


def test_retry(mock_client, mock_response):
    # Setup
    mock_response.with_json({"id": 123, "name": "prkumar"})
    mock_client.with_side_effect([Exception, mock_response])
    github = GitHub(base_url=BASE_URL, client=mock_client)

    # Run
    response = github.get_user("prkumar")

    # Verify
    assert len(mock_client.history) == 2
    assert response.json() == {"id": 123, "name": "prkumar"}


def test_retry_fail(mock_client, mock_response):
    # Setup
    mock_response.with_json({"id": 123, "name": "prkumar"})
    mock_client.with_side_effect([Exception, Exception, mock_response])
    github = GitHub(base_url=BASE_URL, client=mock_client)

    # Run
    with pytest.raises(Exception):
        github.get_issue("prkumar", "uplink", "#1")

    # Verify
    assert len(mock_client.history) == 2


def test_retry_fail_with_client_exception(mock_client, mock_response):
    # Setup
    mock_response.with_json({"id": 123, "name": "prkumar"})
    mock_client.exceptions.ConnectionTimeout = type(
        "ConnectionTimeout", (Exception,), {}
    )
    CustomException = type("CustomException", (Exception,), {})
    mock_client.with_side_effect(
        [
            mock_client.exceptions.ConnectionTimeout,
            CustomException,
            mock_response,
        ]
    )
    github = GitHub(base_url=BASE_URL, client=mock_client)

    # Run
    with pytest.raises(CustomException):
        github.get_project("prkumar", "uplink", "1")

    # Verify
    assert len(mock_client.history) == 2


def test_retry_fail_because_of_wait(mock_client, mock_response):
    # Setup
    mock_response.with_json({"id": 123, "name": "prkumar"})
    mock_client.with_side_effect([Exception, Exception, mock_response])
    github = GitHub(base_url=BASE_URL, client=mock_client)

    # Run
    with pytest.raises(Exception):
        github.get_issue("prkumar", "uplink", "#1")

    # Verify
    assert len(mock_client.history) == 2


def test_retry_with_status_501(mock_client, mock_response):
    # Setup
    mock_response.status_code = 501
    mock_client.with_side_effect([mock_response, Exception])
    github = GitHub(base_url=BASE_URL, client=mock_client)

    # Run
    with pytest.raises(Exception):
        github.get_issues("prkumar", "uplink")

    # Verify
    assert len(mock_client.history) == 2


@pytest_twisted.inlineCallbacks
def test_retry_with_twisted(mock_client, mock_response):
    from twisted.internet import defer

    @defer.inlineCallbacks
    def return_response():
        yield
        defer.returnValue(mock_response)

    # Setup
    mock_response.with_json({"id": 123, "name": "prkumar"})
    mock_client.with_side_effect([Exception, return_response()])
    mock_client.with_io(io.TwistedStrategy())
    github = GitHub(base_url=BASE_URL, client=mock_client)

    # Run
    response = yield github.get_user("prkumar")

    assert len(mock_client.history) == 2
    assert response.json() == {"id": 123, "name": "prkumar"}


@pytest_twisted.inlineCallbacks
def test_retry_fail_with_twisted(mock_client, mock_response):
    from twisted.internet import defer

    @defer.inlineCallbacks
    def return_response():
        yield
        defer.returnValue(mock_response)

    # Setup
    CustomException = type("CustomException", (Exception,), {})
    mock_response.with_json({"id": 123, "name": "prkumar"})
    mock_client.with_side_effect(
        [Exception, CustomException, return_response()]
    )
    mock_client.with_io(io.TwistedStrategy())
    github = GitHub(base_url=BASE_URL, client=mock_client)

    # Run
    with pytest.raises(CustomException):
        yield github.get_user("prkumar")

    assert len(mock_client.history) == 2



# Local imports
from uplink import Consumer, form_url_encoded, put, FieldMap

# Constants
BASE_URL = "https://example.com/"


def test_without_converter(mock_response, mock_client):
    class Calendar(Consumer):
        @form_url_encoded
        @put("/user/repos", args={"event_data": FieldMap})
        def add_event(self, **event_data):
            pass

    mock_client.with_response(mock_response)
    calendar = Calendar(base_url=BASE_URL, client=mock_client)

    # Run
    calendar.add_event(name="Weekly Stand-up", public=True)

    # Assertions: should not convert if converter is None
    request = mock_client.history[0]
    assert request.data == {"name": "Weekly Stand-up", "public": True}



# Standard library imports
import functools

# Local imports
from uplink import converters, decorators, install as _install, returns, utils

__all__ = ["loads", "dumps"]

_get_classes = functools.partial(map, type)


class ResponseBodyConverterFactory(converters.Factory):
    def __init__(self, delegate):
        self.create_response_body_converter = delegate


class RequestBodyConverterFactory(converters.Factory):
    def __init__(self, delegate):
        self.create_request_body_converter = delegate


class _Delegate(object):
    def __init__(self, model_class, annotations, func):
        self._model_class = model_class
        self._annotations = annotations
        self._func = func

    def _contains_annotations(self, argument_annotations, method_annotations):
        types = set(_get_classes(argument_annotations))
        types.update(_get_classes(method_annotations))
        return types.issuperset(self._annotations)

    def _is_relevant(self, type_, request_definition):
        return utils.is_subclass(
            type_, self._model_class
        ) and self._contains_annotations(
            request_definition.argument_annotations,
            request_definition.method_annotations,
        )

    def __call__(self, type_, *args, **kwargs):
        if self._is_relevant(type_, *args, **kwargs):
            return functools.partial(self._func, type_)


class _Wrapper(converters.Factory):
    def __init__(self, w, func):
        self.create_response_body_converter = w.create_response_body_converter
        self.create_request_body_converter = w.create_request_body_converter
        self.create_string_converter = w.create_string_converter
        self._func = func

    def __call__(self, *args, **kwargs):
        return self._func(*args, **kwargs)


class _ModelConverterBuilder(object):
    def __init__(self, base_class, annotations=()):
        """
        Args:
            base_class (type): The base model class.
        """
        self._model_class = base_class
        self._annotations = set(annotations)

    def using(self, func):
        """Sets the converter strategy to the given function."""
        delegate = _Delegate(self._model_class, self._annotations, func)
        return self._wrap_delegate(delegate)

    def _wrap_delegate(self, delegate):  # pragma: no cover
        raise NotImplementedError

    def __call__(self, func):
        converter = _Wrapper(self.using(func), func)
        functools.update_wrapper(converter, func)
        return converter

    install = _install

    @classmethod
    def _make_builder(cls, base_class, annotations, *more_annotations):
        annotations = set(annotations)
        annotations.update(more_annotations)
        return cls(base_class=base_class, annotations=annotations)


# noinspection PyPep8Naming
class loads(_ModelConverterBuilder):
    """
    Builds a custom object deserializer.

    This class takes a single argument, the base model class, and
    registers the decorated function as a deserializer for that base
    class and all subclasses.

    Further, the decorated function should accept two positional
    arguments: (1) the encountered type (which can be the given base
    class or a subclass), and (2) the response data.

    .. code-block:: python

        @loads(ModelBase)
        def load_model(model_cls, data):
            ...

    .. versionadded:: v0.5.0
    """

    def _wrap_delegate(self, delegate):
        return ResponseBodyConverterFactory(delegate)

    @classmethod
    def from_json(cls, base_class, annotations=()):
        """
        Builds a custom JSON deserialization strategy.

        This decorator accepts the same arguments and behaves like
        :py:class:`uplink.loads`, except that the second argument of the
        decorated function is a JSON object:

        .. code-block:: python

            @loads.from_json(User)
            def from_json(user_cls, json):
                return user_cls(json["id"], json["username"])

        Notably, only consumer methods that have the expected return
        type (i.e., the given base class or any subclass) and are
        decorated with :py:class:`uplink.returns.from_json` can leverage
        the registered strategy to deserialize JSON responses.

        For example, the following consumer method would leverage the
        :py:func:`from_json` strategy defined above:

        .. code-block:: python

            @returns.from_json
            @get("user")
            def get_user(self) -> User: pass

        .. versionadded:: v0.5.0
        """
        return cls._make_builder(base_class, annotations, returns.json)


# noinspection PyPep8Naming
class dumps(_ModelConverterBuilder):
    """
    Builds a custom object serializer.

    This decorator takes a single argument, the base model class, and
    registers the decorated function as a serializer for that base
    class and all subclasses.

    Further, the decorated function should accept two positional
    arguments: (1) the encountered type (which can be the given base
    class or a subclass), and (2) the encountered instance.

    .. code-block:: python

        @dumps(ModelBase)
        def deserialize_model(model_cls, model_instance):
            ...

    .. versionadded:: v0.5.0
    """

    def _wrap_delegate(self, delegate):
        return RequestBodyConverterFactory(delegate)

    @classmethod
    def to_json(cls, base_class, annotations=()):
        """
        Builds a custom JSON serialization strategy.

        This decorator accepts the same arguments and behaves like
        :py:class:`uplink.dumps`. The only distinction is that the
        decorated function should be JSON serializable.

        .. code-block:: python

            @dumps.to_json(ModelBase)
            def to_json(model_cls, model_instance):
                return model_instance.to_json()

        Notably, only consumer methods that are decorated with
        py:class:`uplink.json` and have one or more argument annotations
        with the expected type (i.e., the given base class or a subclass)
        can leverage the registered strategy.

        For example, the following consumer method would leverage the
        :py:func:`to_json` strategy defined above, given
        :py:class:`User` is a subclass of :py:class:`ModelBase`:

        .. code-block:: python

            @json
            @post("user")
            def change_user_name(self, name: Field(type=User): pass

        .. versionadded:: v0.5.0
        """
        return cls._make_builder(base_class, annotations, decorators.json)



# Third-party imports
import pytest

# Local imports
from uplink import auth
from uplink import utils


class TestGetAuth(object):
    def test_none(self):
        authentication = auth.get_auth(None)
        assert authentication == utils.no_op

    def test_tuple(self):
        authentication = auth.get_auth(("username", "password"))
        assert isinstance(authentication, auth.BasicAuth)

    def test_callable(self):
        def func():
            pass

        output = auth.get_auth(func)
        assert output is func

    def test_unsupported(self):
        with pytest.raises(ValueError):
            auth.get_auth(object())


def test_api_token_param(request_builder):
    # Setup
    token_param_auth = auth.ApiTokenParam(param="token-param", token="token-value")

    # Verify
    token_param_auth(request_builder)
    assert request_builder.info["params"]["token-param"] == "token-value"


def test_api_token_header_without_prefix(request_builder):
    # Setup
    token_header_auth = auth.ApiTokenHeader("Token-Header", "token-value")

    # Verify
    token_header_auth(request_builder)
    assert request_builder.info["headers"]["Token-Header"] == "token-value"


def test_api_token_header_with_prefix(request_builder):
    # Setup
    token_header_auth = auth.ApiTokenHeader("Token-Header", "token-value", prefix="Prefix")

    # Verify
    token_header_auth(request_builder)
    assert request_builder.info["headers"]["Token-Header"] == "Prefix token-value"


def test_api_token_header_subclass_without_prefix(request_builder):
    # Setup

    class ApiTokenHeaderSubclass(auth.ApiTokenHeader):
        _header = "Token-Header"

        def __init__(self, token):
            self._token = token

    token_header_auth = ApiTokenHeaderSubclass("token-value")

    # Verify
    token_header_auth(request_builder)
    assert request_builder.info["headers"]["Token-Header"] == "token-value"


def test_api_token_header_subclass_with_prefix(request_builder):
    # Setup

    class ApiTokenHeaderSubclass(auth.ApiTokenHeader):
        _header = "Token-Header"
        _prefix = "Prefix"

        def __init__(self, token):
            self._token = token

    token_header_auth = ApiTokenHeaderSubclass("token-value")

    # Verify
    token_header_auth(request_builder)
    assert request_builder.info["headers"]["Token-Header"] == "Prefix token-value"


def test_basic_auth(request_builder):
    # Setup
    basic_auth = auth.BasicAuth("username", "password")

    # Verify
    basic_auth(request_builder)
    auth_str = basic_auth._header_value
    assert request_builder.info["headers"]["Authorization"] == auth_str


def test_proxy_auth(request_builder):
    # Setup
    proxy_auth = auth.ProxyAuth("username", "password")

    # Verify
    proxy_auth(request_builder)
    auth_str = proxy_auth._header_value
    assert request_builder.info["headers"]["Proxy-Authorization"] == auth_str


def test_bearer_token(request_builder):
    # Setup
    bearer_token = auth.BearerToken("bearer-token")

    # Verify
    bearer_token(request_builder)
    auth_str = bearer_token._header_value
    assert request_builder.info["headers"]["Authorization"] == auth_str


class TestMultiAuth(object):
    def setup_basic_auth(self):
        return auth.BasicAuth("apiuser", "apipass")

    def verify_basic_auth(self, basic_auth, request_builder):
        basic_auth_str = basic_auth._header_value
        assert request_builder.info["headers"]["Authorization"] == basic_auth_str

    def setup_proxy_auth(self):
        return auth.ProxyAuth("proxyuser", "proxypass")

    def verify_proxy_auth(self, proxy_auth, request_builder):
        proxy_auth_str = proxy_auth._header_value
        assert request_builder.info["headers"]["Proxy-Authorization"] == proxy_auth_str

    def setup_param_auth(self):
        return auth.ApiTokenParam(param="token-param", token="token-value")

    def verify_param_auth(self, request_builder):
        assert request_builder.info["params"]["token-param"] == "token-value"

    def test_len(self):
        multi_auth = auth.MultiAuth()
        assert len(multi_auth) == 0

    def test_none(self):
        multi_auth = auth.MultiAuth(None)
        assert len(multi_auth) == 1
        assert multi_auth[0] == utils.no_op

    def test_one_method(self, request_builder):
        # Setup
        basic_auth = self.setup_basic_auth()
        multi_auth = auth.MultiAuth(basic_auth)

        # Verify
        assert len(multi_auth) == 1
        assert multi_auth[0] == basic_auth

        multi_auth(request_builder)
        self.verify_basic_auth(basic_auth, request_builder)

    def test_four_methods(self, request_builder):
        # Setup
        param_auth = self.setup_param_auth()
        basic_auth = self.setup_basic_auth()
        proxy_auth = self.setup_proxy_auth()
        multi_auth = auth.MultiAuth(None, param_auth, basic_auth, proxy_auth)

        # Verify
        assert len(multi_auth) == 4
        assert multi_auth[0] == utils.no_op
        assert multi_auth[1] == param_auth
        assert multi_auth[2] == basic_auth
        assert multi_auth[3] == proxy_auth

        multi_auth(request_builder)
        self.verify_param_auth(request_builder)
        self.verify_basic_auth(basic_auth, request_builder)
        self.verify_proxy_auth(proxy_auth, request_builder)

    def test_append(self, request_builder):
        # Setup
        basic_auth = self.setup_basic_auth()
        proxy_auth = self.setup_proxy_auth()
        multi_auth = auth.MultiAuth()
        multi_auth.append(basic_auth)
        multi_auth.append(proxy_auth)

        # Verify
        assert len(multi_auth) == 2
        assert multi_auth[0] == basic_auth
        assert multi_auth[1] == proxy_auth

        multi_auth(request_builder)
        self.verify_basic_auth(basic_auth, request_builder)
        self.verify_proxy_auth(proxy_auth, request_builder)

    def test_extend(self, request_builder):
        # Setup
        basic_auth = self.setup_basic_auth()
        proxy_auth = self.setup_proxy_auth()
        multi_auth = auth.MultiAuth()
        multi_auth.extend([basic_auth, proxy_auth])

        # Verify
        assert len(multi_auth) == 2
        assert multi_auth[0] == basic_auth
        assert multi_auth[1] == proxy_auth

        multi_auth(request_builder)
        self.verify_basic_auth(basic_auth, request_builder)
        self.verify_proxy_auth(proxy_auth, request_builder)



"""
This module defines a converter that uses :py:mod:`marshmallow` schemas
to deserialize and serialize values.
"""

# Local imports
from uplink import utils
from uplink.converters import interfaces, register_default_converter_factory


class MarshmallowConverter(interfaces.Factory):
    """
    A converter that serializes and deserializes values using
    :py:mod:`marshmallow` schemas.

    To deserialize JSON responses into Python objects with this
    converter, define a :py:class:`marshmallow.Schema` subclass and set
    it as the return annotation of a consumer method:

    .. code-block:: python

        @get("/users")
        def get_users(self, username) -> UserSchema():
            '''Fetch a single user'''

    Note:

        This converter is an optional feature and requires the
        :py:mod:`marshmallow` package. For example, here's how to
        install this feature using pip::

            $ pip install uplink[marshmallow]
    """

    try:
        import marshmallow
    except ImportError:  # pragma: no cover
        marshmallow = None
        is_marshmallow_3 = None
    else:
        is_marshmallow_3 = marshmallow.__version__ >= "3.0"

    def __init__(self):
        if self.marshmallow is None:
            raise ImportError("No module named 'marshmallow'")

    class ResponseBodyConverter(interfaces.Converter):
        def __init__(self, extract_data, schema):
            self._extract_data = extract_data
            self._schema = schema

        def convert(self, response):
            try:
                json = response.json()
            except AttributeError:
                # Assume that the response is already json
                json = response

            return self._extract_data(self._schema.load(json))

    class RequestBodyConverter(interfaces.Converter):
        def __init__(self, extract_data, schema):
            self._extract_data = extract_data
            self._schema = schema

        def convert(self, value):
            return self._extract_data(self._schema.dump(value))

    @classmethod
    def _get_schema(cls, type_):
        if utils.is_subclass(type_, cls.marshmallow.Schema):
            return type_()
        elif isinstance(type_, cls.marshmallow.Schema):
            return type_
        raise ValueError("Expected marshmallow.Scheme subclass or instance.")

    def _extract_data(self, m):
        # After marshmallow 3.0, Schema.load() and Schema.dump() don't
        # return a (data, errors) tuple any more. Only `data` is returned.
        return m if self.is_marshmallow_3 else m.data

    def _make_converter(self, converter_cls, type_):
        try:
            # Try to generate schema instance from the given type.
            schema = self._get_schema(type_)
        except ValueError:
            # Failure: the given type is not a `marshmallow.Schema`.
            return None
        else:
            return converter_cls(self._extract_data, schema)

    def create_request_body_converter(self, type_, *args, **kwargs):
        return self._make_converter(self.RequestBodyConverter, type_)

    def create_response_body_converter(self, type_, *args, **kwargs):
        return self._make_converter(self.ResponseBodyConverter, type_)

    @classmethod
    def register_if_necessary(cls, register_func):
        if cls.marshmallow is not None:
            register_func(cls)


MarshmallowConverter.register_if_necessary(register_default_converter_factory)



# Standard library imports
import operator

# Local imports
from uplink.clients.io import RequestTemplate, transitions


class DefaultRequestTemplate(RequestTemplate):
    """The fallback behaviors for all hooks."""

    def before_request(self, request):
        return transitions.send(request)

    def after_response(self, request, response):
        return transitions.finish(response)

    def after_exception(self, request, exc_type, exc_val, exc_tb):
        return transitions.fail(exc_type, exc_val, exc_tb)


class CompositeRequestTemplate(RequestTemplate):
    """A chain of many templates with fallback behaviors."""

    __FALLBACK = DefaultRequestTemplate()

    def _get_transition(self, method, *args, **kwargs):
        caller = operator.methodcaller(method, *args, **kwargs)
        for template in self._templates:
            transition = caller(template)
            if transition is not None:
                return transition
        else:
            return caller(self._fallback)

    def __init__(self, templates, fallback=__FALLBACK):
        self._templates = list(templates)
        self._fallback = fallback

    def before_request(self, request):
        return self._get_transition(
            RequestTemplate.before_request.__name__, request
        )

    def after_response(self, request, response):
        return self._get_transition(
            RequestTemplate.after_response.__name__, request, response
        )

    def after_exception(self, request, exc_type, exc_val, exc_tb):
        return self._get_transition(
            RequestTemplate.after_exception.__name__,
            request,
            exc_type,
            exc_val,
            exc_tb,
        )



# Standard library imports
import sys
import warnings

# Local imports
from uplink import decorators
from uplink.converters import keys, interfaces

__all__ = ["json", "from_json", "schema"]


class ReturnType(object):
    def __init__(self, decorator, type_):
        self._decorator = decorator
        self._type_ = type_

    @property
    def type(self):
        return self._type_

    @staticmethod
    def with_decorator(old, decorator):
        old_type = None if old is None else old.type
        return ReturnType(decorator, decorator.return_type or old_type)

    def with_strategy(self, strategy):
        return CallableReturnType(self._decorator, self._type_, strategy)

    def is_applicable(self, decorator):
        return self._decorator is decorator


class CallableReturnType(ReturnType):
    def __init__(self, decorator, type_, strategy):
        super(CallableReturnType, self).__init__(decorator, type_)
        self._strategy = strategy

    def __call__(self, *args, **kwargs):
        return self._strategy(*args, **kwargs)


class _ReturnsBase(decorators.MethodAnnotation):
    @property
    def return_type(self):  # pragma: no cover
        raise NotImplementedError

    def _make_strategy(self, converter):  # pragma: no cover
        pass

    def _modify_request_definition(self, definition, kwargs):
        super(_ReturnsBase, self)._modify_request_definition(definition, kwargs)
        definition.return_type = ReturnType.with_decorator(
            definition.return_type, self
        )

    def _get_converter(self, request_builder, return_type):  # pragma: no cover
        return request_builder.get_converter(
            keys.CONVERT_FROM_RESPONSE_BODY, return_type.type
        )

    def modify_request(self, request_builder):
        return_type = request_builder.return_type
        if not return_type.is_applicable(self):
            return

        converter = self._get_converter(request_builder, return_type)
        if converter is None:
            return

        # Found a converter that can handle the return type.
        request_builder.return_type = return_type.with_strategy(
            self._make_strategy(converter)
        )


class JsonStrategy(object):
    # TODO: Consider moving this under json decorator
    # TODO: Support JSON Pointer (https://tools.ietf.org/html/rfc6901)

    def __init__(self, converter, key=()):
        self._converter = converter

        if not isinstance(key, (list, tuple)):
            key = (key,)
        self._key = key

    def __call__(self, response):
        content = response.json()
        for name in self._key:
            content = content[name]
        content = self._converter(content)
        return content


# noinspection PyPep8Naming
class json(_ReturnsBase):
    """
    Specifies that the decorated consumer method should return a JSON
    object.

    .. code-block:: python

        # This method will return a JSON object (e.g., a dict or list)
        @returns.json
        @get("/users/{username}")
        def get_user(self, username):
            \"""Get a specific user.\"""

    Returning a Specific JSON Field:

        The :py:attr:`key` argument accepts a string or tuple that
        specifies the path of an internal field in the JSON document.

        For instance, consider an API that returns JSON responses that,
        at the root of the document, contains both the server-retrieved
        data and a list of relevant API errors:

        .. code-block:: json
            :emphasize-lines: 2

            {
                "data": { "user": "prkumar", "id": "140232" },
                "errors": []
            }

        If returning the list of errors is unnecessary, we can use the
        :py:attr:`key` argument to strictly return the nested field
        :py:attr:`data.id`:

        .. code-block:: python

            @returns.json(key=("data", "id"))
            @get("/users/{username}")
            def get_user_id(self, username):
                \"""Get a specific user's ID.\"""

        We can also configure Uplink to convert the field before it's
        returned by also specifying the``type`` argument:

       .. code-block:: python

            @returns.json(key=("data", "id"), type=int)
            @get("/users/{username}")
            def get_user_id(self, username):
                \"""Get a specific user's ID.\"""

    .. versionadded:: v0.5.0
    """

    _can_be_static = True

    class _DummyConverter(interfaces.Converter):
        def convert(self, response):
            return response

    class _CastConverter(interfaces.Converter):
        def __init__(self, cast):
            self._cast = cast

        def convert(self, response):
            return self._cast(response)

    __dummy_converter = _DummyConverter()

    def __init__(self, type=None, key=(), model=None, member=()):
        if model:  # pragma: no cover
            warnings.warn(
                "The `model` argument of @returns.json is deprecated and will "
                "be removed in v1.0.0. Use `type` instead.",
                DeprecationWarning,
            )
        if member:  # pragma: no cover
            warnings.warn(
                "The `member` argument of @returns.json is deprecated and will "
                "be removed in v1.0.0. Use `key` instead.",
                DeprecationWarning,
            )
        self._type = type or model
        self._key = key or member

    @property
    def return_type(self):
        return self._type

    def _get_converter(self, request_builder, return_type):
        converter = super(json, self)._get_converter(
            request_builder, return_type
        )

        if converter:
            return converter

        if callable(return_type.type):
            return self._CastConverter(return_type.type)

        # If the return_type cannot be converted, the strategy should directly
        # return the JSON body of the HTTP response, instead of trying to
        # deserialize it into a certain type. In this case, by
        # defaulting the return type to the dummy converter, which
        # implements this pass-through behavior, we ensure that
        # _make_strategy is called.
        return self.__dummy_converter

    def _make_strategy(self, converter):
        return JsonStrategy(converter, self._key)


from_json = json
"""
    Specifies that the decorated consumer method should produce
    instances of a :py:obj:`type` class using a registered
    deserialization strategy (see :py:meth:`uplink.loads.from_json`)

    This decorator accepts the same arguments as
    :py:class:`uplink.returns.json`.

    Often, a JSON response body represents a schema in your application.
    If an existing Python object encapsulates this schema, use the
    :py:attr:`type` argument to specify it as the return type:

    .. code-block:: python

        @returns.from_json(type=User)
        @get("/users/{username}")
        def get_user(self, username):
            \"""Get a specific user.\"""

    For Python 3 users, you can alternatively provide a return value
    annotation. Hence, the previous code is equivalent to the following
    in Python 3:

    .. code-block:: python

        @returns.from_json
        @get("/users/{username}")
        def get_user(self, username) -> User:
            \"""Get a specific user.\"""

    Both usages typically require also registering a converter that
    knows how to deserialize the JSON into the specified :py:attr:`type`
    (see :py:meth:`uplink.loads.from_json`). This step is unnecessary if
    the :py:attr:`type` is defined using a library for which Uplink has
    built-in support, such as :py:mod:`marshmallow`.

    .. versionadded:: v0.6.0
"""


# noinspection PyPep8Naming
class schema(_ReturnsBase):
    """
    Specifies that the function returns a specific type of response.

    In Python 3, to provide a consumer method's return type, you can
    set it as the method's return annotation:

    .. code-block:: python

        @get("/users/{username}")
        def get_user(self, username) -> UserSchema:
            \"""Get a specific user.\"""

    For Python 2.7 compatibility, you can use this decorator instead:

    .. code-block:: python

        @returns.schema(UserSchema)
        @get("/users/{username}")
        def get_user(self, username):
            \"""Get a specific user.\"""

    To have Uplink convert response bodies into the desired type, you
    will need to define an appropriate converter (e.g., using
    :py:class:`uplink.loads`).

    .. versionadded:: v0.5.1
    """

    def __init__(self, type):
        self._schema = type

    @property
    def return_type(self):
        return self._schema

    def _make_strategy(self, converter):
        return converter


class _ModuleProxy(object):
    __module = sys.modules[__name__]

    schema = model = schema
    json = json
    from_json = from_json
    __all__ = __module.__all__

    def __getattr__(self, item):
        return getattr(self.__module, item)

    def __call__(self, *args, **kwargs):
        return schema(*args, **kwargs)


sys.modules[__name__] = _ModuleProxy()



"""
This module provides a class for defining custom handling for specific
points of an HTTP transaction.
"""
# Local imports
from uplink import compat

__all__ = ["TransactionHook", "RequestAuditor", "ResponseHandler"]


def _wrap_if_necessary(hook, requires_consumer):
    if not requires_consumer:
        return _wrap_to_ignore_consumer(hook)
    return hook


def _wrap_to_ignore_consumer(hook):
    @compat.wraps(hook)
    def wrapper(_, *args, **kwargs):
        # Expects that consumer is the first argument
        return hook(*args, **kwargs)

    return wrapper


class TransactionHook(object):
    """
    A utility class providing methods that define hooks for specific
    points of an HTTP transaction.
    """

    def audit_request(self, consumer, request_builder):  # pragma: no cover
        """Inspects details of a request before it is sent."""
        pass

    handle_response = None
    """
    Handles a response object from the server.

    This method can be undefined (i.e., None), indicating that this hook
    does not handle responses.

    Args:
        response: The received HTTP response.
    """

    def handle_exception(
        self, consumer, exc_type, exc_val, exc_tb
    ):  # pragma: no cover
        """
        Handles an exception thrown while waiting for a response from
        the server.

        Args:
            consumer: The consumer that spawned the failing request.
            exc_type: The type of the exception.
            exc_val: The exception instance raised.
            exc_tb: A traceback instance.
        """
        pass


class TransactionHookChain(TransactionHook):
    """
    A chain that conjoins several transaction hooks into a single
    object.

    A method call on this composite object invokes the corresponding
    method on all hooks in the chain.
    """

    def __init__(self, *hooks):
        self._hooks = hooks
        self._response_handlers = []

        # TODO: If more than one callback exists on the chain, the chain
        # expects it can execute each synchronously. Instead, we should
        # be smart about this and produces a chained coroutine when all
        # callbacks are coroutines, so that the client can execute the
        # chain asynchronously. Further, when provided both synchronous
        # and asynchronous callbacks, we should raise an exception when
        # the order is mixed and split into two chains (one async and
        # the other sync) when the order permits separation.

        # Adding a synchronous callback to an async request forces the
        # request to execute synchronously while running this chain. To
        # avoid unnecessarily executing this chain when no callbacks
        # exists, we can set the `handle_response` method to null,
        # indicating that this hook doesn't handle responses.
        response_handlers = [h for h in hooks if h.handle_response is not None]
        if not response_handlers:
            self.handle_response = None
        elif len(response_handlers) == 1:
            self.handle_response = response_handlers[0].handle_response

        self._response_handlers = response_handlers

    def audit_request(self, consumer, request_handler):
        for hook in self._hooks:
            hook.audit_request(consumer, request_handler)

    def handle_response(self, consumer, response):
        for hook in self._response_handlers:
            response = hook.handle_response(consumer, response)
        return response

    def handle_exception(self, consumer, exc_type, exc_val, exc_tb):
        for hook in self._hooks:
            hook.handle_exception(consumer, exc_type, exc_val, exc_tb)
        compat.reraise(exc_type, exc_val, exc_tb)


class RequestAuditor(TransactionHook):
    """
    Transaction hook that inspects requests using a function provided at
    time of instantiation.
    """

    def __init__(self, auditor, requires_consumer=False):
        self.audit_request = _wrap_if_necessary(auditor, requires_consumer)


class ResponseHandler(TransactionHook):
    """
    Transaction hook that handles responses using a function provided at
    time of instantiation.
    """

    def __init__(self, handler, requires_consumer=False):
        self.handle_response = _wrap_if_necessary(handler, requires_consumer)


class ExceptionHandler(TransactionHook):
    """
    Transaction hook that handles an exception thrown while waiting for
    a response, using the provided function.
    """

    def __init__(self, exception_handler, requires_consumer=False):
        self.handle_exception = _wrap_if_necessary(
            exception_handler, requires_consumer
        )



"""
This module defines common converter keys, used by consumers of the
converter layer to identify the desired conversion type when querying a
:py:class:`uplink.converters.ConverterFactoryRegistry`.
"""
# Standard library imports
import functools

# Local imports

__all__ = [
    "CONVERT_TO_STRING",
    "CONVERT_TO_REQUEST_BODY",
    "CONVERT_FROM_RESPONSE_BODY",
    "Map",
    "Sequence",
]

#: Object to string conversion.
CONVERT_TO_STRING = 0

#: Object to request body conversion.
CONVERT_TO_REQUEST_BODY = 1

# Response body to object conversion.
CONVERT_FROM_RESPONSE_BODY = 2


class CompositeKey(object):
    """
    A utility class for defining composable converter keys.

    Arguments:
        converter_key: The enveloped converter key.
    """

    def __init__(self, converter_key):
        self._converter_key = converter_key

    def __eq__(self, other):
        if isinstance(other, CompositeKey) and type(other) is type(self):
            return other._converter_key == self._converter_key
        return False

    def convert(self, converter, value):  # pragma: no cover
        raise NotImplementedError

    def __call__(self, converter_registry):
        factory = converter_registry[self._converter_key]

        def factory_wrapper(*args, **kwargs):
            converter = factory(*args, **kwargs)
            if not converter:
                return None
            return functools.partial(self.convert, converter)

        return factory_wrapper


class Map(CompositeKey):
    """
    Object to mapping conversion.

    The constructor argument :py:data:`converter_key` details the type
    for the values in the mapping. For instance::

        # Key for conversion of an object to a mapping of strings.
        Map(CONVERT_TO_STRING)
    """

    def convert(self, converter, value):
        return dict((k, converter(value[k])) for k in value)


class Sequence(CompositeKey):
    """
    Object to sequence conversion.

    The constructor argument :py:data:`converter_key` details the type
    for the elements in the sequence. For instance::

        # Key for conversion of an object to a sequence of strings.
        Sequence(CONVERT_TO_STRING)
    """

    def convert(self, converter, value):
        if isinstance(value, (list, tuple)):
            return list(map(converter, value))
        else:
            return converter(value)


class Identity(object):
    """
    Identity conversion - pass value as is
    """

    def __call__(self, converter_registry):
        return self._identity_factory

    def __eq__(self, other):
        return type(other) is type(self)

    def _identity_factory(self, *args, **kwargs):
        return self._identity

    def _identity(self, value):
        return value



# Local imports.
import uplink

# Constants
BASE_URL = "https://example.com/"


@uplink.response_handler(requires_consumer=True)
def handle_response_with_consumer(consumer, response):
    consumer.flagged = True
    return response


@uplink.response_handler
def handle_response(response):
    response.flagged = True
    return response


class WrappedException(Exception):
    def __init__(self, exception):
        self.exception = exception


@uplink.error_handler(requires_consumer=True)
def handle_error_with_consumer(consumer, exc_type, exc_value, exc_tb):
    consumer.flagged = True
    raise WrappedException(exc_value)


@uplink.error_handler
def handle_error(exc_type, exc_value, exc_tb):
    raise WrappedException(exc_value)


class Calendar(uplink.Consumer):
    @handle_response_with_consumer
    @uplink.get("todos/{todo_id}")
    def get_todo(self, todo_id):
        pass

    @handle_response
    @uplink.get("months/{name}/todos")
    def get_month(self, name):
        pass

    @handle_response_with_consumer
    @handle_response
    @uplink.get("months/{month}/days/{day}/todos")
    def get_day(self, month, day):
        pass

    @handle_error_with_consumer
    @uplink.get("users/{user_id}")
    def get_user(self, user_id):
        pass

    @handle_error
    @uplink.get("events/{event_id}")
    def get_event(self, event_id):
        pass


@handle_response
class CalendarV2(uplink.Consumer):
    @uplink.get("todos/{todo_id}")
    def get_todo(self, todo_id):
        pass


def test_class_response_handler(mock_client):
    calendar = CalendarV2(base_url=BASE_URL, client=mock_client)
    calendar.flagged = False

    # Run
    response = calendar.get_todo(todo_id=1)

    # Verify
    assert response.flagged is True


def test_response_handler_with_consumer(mock_client):
    calendar = Calendar(base_url=BASE_URL, client=mock_client)
    calendar.flagged = False

    # Run
    calendar.get_todo(todo_id=1)

    # Verify
    assert calendar.flagged is True


def test_response_handler(mock_client):
    calendar = Calendar(base_url=BASE_URL, client=mock_client)

    # Run
    response = calendar.get_month("September")

    # Verify
    assert response.flagged is True


def test_multiple_response_handlers(mock_client):
    calendar = Calendar(base_url=BASE_URL, client=mock_client)

    # Run
    response = calendar.get_day("September", 2)

    # Verify
    assert response.flagged
    assert calendar.flagged


def test_error_handler_with_consumer(mock_client):
    # Setup: raise specific exception
    expected_error = IOError()
    mock_client.with_side_effect(expected_error)

    calendar = Calendar(base_url=BASE_URL, client=mock_client)
    calendar.flagged = False

    # Run
    try:
        calendar.get_user(user_id=1)
    except WrappedException as err:
        assert err.exception == expected_error
        assert calendar.flagged is True
    else:
        raise AssertionError


def test_error_handler(mock_client):
    # Setup: raise specific exception
    expected_error = IOError()
    mock_client.with_side_effect(expected_error)

    calendar = Calendar(base_url=BASE_URL, client=mock_client)

    # Run
    try:
        calendar.get_event(event_id=1)
    except WrappedException as err:
        assert err.exception == expected_error
    else:
        raise AssertionError



"""
This module defines a converter that uses :py:mod:`pydantic` models
to deserialize and serialize values.
"""

from uplink.converters import register_default_converter_factory
from uplink.converters.interfaces import Factory, Converter
from uplink.utils import is_subclass


def _encode_pydantic(obj):
    from pydantic.json import pydantic_encoder

    # json atoms
    if isinstance(obj, (str, int, float, bool)) or obj is None:
        return obj

    # json containers
    if isinstance(obj, dict):
        return {_encode_pydantic(k): _encode_pydantic(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [_encode_pydantic(i) for i in obj]

    # pydantic types
    return _encode_pydantic(pydantic_encoder(obj))


class _PydanticRequestBody(Converter):
    def __init__(self, model):
        self._model = model

    def convert(self, value):
        if isinstance(value, self._model):
            return _encode_pydantic(value)
        return _encode_pydantic(self._model.parse_obj(value))


class _PydanticResponseBody(Converter):
    def __init__(self, model):
        self._model = model

    def convert(self, response):
        try:
            data = response.json()
        except AttributeError:
            data = response

        return self._model.parse_obj(data)


class PydanticConverter(Factory):
    """
    A converter that serializes and deserializes values using
    :py:mod:`pydantic` models.

    To deserialize JSON responses into Python objects with this
    converter, define a :py:class:`pydantic.BaseModel` subclass and set
    it as the return annotation of a consumer method:

    .. code-block:: python

        @returns.json()
        @get("/users")
        def get_users(self, username) -> List[UserModel]:
            '''Fetch multiple users'''

    Note:

        This converter is an optional feature and requires the
        :py:mod:`pydantic` package. For example, here's how to
        install this feature using pip::

            $ pip install uplink[pydantic]
    """

    try:
        import pydantic
    except ImportError:  # pragma: no cover
        pydantic = None

    def __init__(self):
        """
        Validates if :py:mod:`pydantic` is installed
        """
        if self.pydantic is None:
            raise ImportError("No module named 'pydantic'")

    def _get_model(self, type_):
        if is_subclass(type_, self.pydantic.BaseModel):
            return type_
        raise ValueError("Expected pydantic.BaseModel subclass or instance")

    def _make_converter(self, converter, type_):
        try:
            model = self._get_model(type_)
        except ValueError:
            return None

        return converter(model)

    def create_request_body_converter(self, type_, *args, **kwargs):
        return self._make_converter(_PydanticRequestBody, type_)

    def create_response_body_converter(self, type_, *args, **kwargs):
        return self._make_converter(_PydanticResponseBody, type_)

    @classmethod
    def register_if_necessary(cls, register_func):
        if cls.pydantic is not None:
            register_func(cls)


PydanticConverter.register_if_necessary(register_default_converter_factory)



# Standard library imports
import sys

# Third-party imports
from twisted.internet import reactor, defer, task

# Local imports
from uplink.clients.io import interfaces


__all__ = ["TwistedStrategy"]


class TwistedStrategy(interfaces.IOStrategy):
    """A non-blocking execution strategy using asyncio."""

    _deferred = None

    @defer.inlineCallbacks
    def invoke(self, func, args, kwargs, callback):
        try:
            response = yield func(*args, **kwargs)
        except Exception as error:
            tb = sys.exc_info()[2]
            response = yield callback.on_failure(type(error), error, tb)
        else:
            response = yield callback.on_success(response)
        defer.returnValue(response)

    @defer.inlineCallbacks
    def sleep(self, duration, callback):
        yield task.deferLater(reactor, duration, lambda: None)
        response = yield callback.on_success()
        defer.returnValue(response)

    @defer.inlineCallbacks
    def finish(self, response):
        yield
        defer.returnValue(response)

    @defer.inlineCallbacks
    def fail(self, exc_type, exc_val, exc_tb):
        yield
        super(TwistedStrategy, self).fail(exc_type, exc_val, exc_tb)

    @defer.inlineCallbacks
    def execute(self, executable):
        response = yield executable.execute()
        defer.returnValue(response)



# Standard library imports

# Third party imports
import requests

# Local imports
from uplink.clients import exceptions, io, interfaces, register


class RequestsClient(interfaces.HttpClientAdapter):
    """
    A :py:mod:`requests` client that returns
    :py:class:`requests.Response` responses.

    Args:
        session (:py:class:`requests.Session`, optional): The session
            that should handle sending requests. If this argument is
            omitted or set to :py:obj:`None`, a new session will be
            created.
    """

    exceptions = exceptions.Exceptions()

    def __init__(self, session=None, **kwargs):
        self.__auto_created_session = False
        if session is None:
            session = self._create_session(**kwargs)
            self.__auto_created_session = True
        self.__session = session

    def __del__(self):
        if self.__auto_created_session:
            self.__session.close()

    @staticmethod
    @register.handler
    def with_session(session, *args, **kwargs):
        if isinstance(session, requests.Session):
            return RequestsClient(session, *args, **kwargs)

    @staticmethod
    def _create_session(**kwargs):
        session = requests.Session()
        for key in kwargs:
            setattr(session, key, kwargs[key])
        return session

    def send(self, request):
        method, url, extras = request
        return self.__session.request(method=method, url=url, **extras)

    def apply_callback(self, callback, response):
        return callback(response)

    @staticmethod
    def io():
        return io.BlockingStrategy()


# === Register client exceptions === #
RequestsClient.exceptions.BaseClientException = requests.RequestException
RequestsClient.exceptions.ConnectionError = requests.ConnectionError
RequestsClient.exceptions.ConnectionTimeout = requests.ConnectTimeout
RequestsClient.exceptions.ServerTimeout = requests.ReadTimeout
RequestsClient.exceptions.SSLError = requests.exceptions.SSLError
RequestsClient.exceptions.InvalidURL = requests.exceptions.InvalidURL




__all__ = ["after_attempt", "after_delay"]


class RetryBreaker(object):
    def __or__(self, other):
        if other is not None:
            assert isinstance(
                other, RetryBreaker
            ), "Both objects should be retry breakers."
            return _Or(self, other)
        return self

    def __call__(self):  # pragma: no cover
        raise NotImplementedError


class _Or(RetryBreaker):
    def __init__(self, left, right):
        self._left = left
        self._right = right

    def __call__(self):
        left = self._left()
        right = self._right()
        while True:
            delay = yield
            next(left)
            next(right)
            stop_left = left.send(delay)
            stop_right = right.send(delay)
            yield stop_left or stop_right


# noinspection PyPep8Naming
class after_attempt(RetryBreaker):
    """Stops retrying after the specified number of ``attempts``."""

    def __init__(self, attempt):
        self._max_attempt = attempt
        self._attempt = 0

    def __call__(self):
        attempt = 0
        while True:
            yield
            attempt += 1
            yield self._max_attempt <= attempt


# noinspection PyPep8Naming
class after_delay(RetryBreaker):
    """
    Stops retrying after the backoff exceeds the specified ``delay``
    in seconds.
    """

    def __init__(self, delay):
        self._max_delay = delay

    def __call__(self):
        while True:
            delay = yield
            yield self._max_delay < delay


class _NeverStop(RetryBreaker):
    def __call__(self):
        while True:
            yield
            yield False


#: Continuously retry until the server returns a successful response
NEVER = _NeverStop()

# Keep for backwards compatibility with v0.8.0
# TODO: Remove in v1.0.0
DISABLE = NEVER



class ClientExceptionProxy(object):
    def __init__(self, getter):
        self._getter = getter

    @classmethod
    def wrap_proxy_if_necessary(cls, exc):
        return exc if isinstance(exc, cls) else (lambda exceptions: exc)

    def __call__(self, exceptions):
        return self._getter(exceptions)



"""
This module defines an :py:class:`aiohttp.ClientSession` adapter
that returns awaitable responses.
"""
# Standard library imports
import asyncio
import collections
import inspect
import threading
from concurrent import futures

# Third-party imports
try:
    import aiohttp
except ImportError:  # pragma: no cover
    aiohttp = None

# Local imports
from uplink.clients import exceptions, io, interfaces, register


def threaded_callback(callback):
    async def new_callback(response):
        if isinstance(response, aiohttp.ClientResponse):
            await response.text()
            response = ThreadedResponse(response)
        response = callback(response)
        if isinstance(response, ThreadedResponse):
            return response.unwrap()
        else:
            return response

    return new_callback


class AiohttpClient(interfaces.HttpClientAdapter):
    """
    An :py:mod:`aiohttp` client that creates awaitable responses.

    Note:
        This client is an optional feature and requires the :py:mod:`aiohttp`
        package. For example, here's how to install this extra using pip::

            $ pip install uplink[aiohttp]

    Args:
        session (:py:class:`aiohttp.ClientSession`, optional):
            The session that should handle sending requests. If this
            argument is omitted or set to :py:obj:`None`, a new session
            will be created.
    """

    exceptions = exceptions.Exceptions()

    # TODO: Update docstrings to include aiohttp constructor parameters.

    __ARG_SPEC = collections.namedtuple("__ARG_SPEC", "args kwargs")

    def __init__(self, session=None, **kwargs):
        if aiohttp is None:
            raise NotImplementedError("aiohttp is not installed.")
        self._auto_created_session = False
        if session is None:
            session = self._create_session(**kwargs)
        self._session = session
        self._sync_callback_adapter = threaded_callback

    def __del__(self):
        # TODO: Consider replacing this with a close method
        if self._auto_created_session:
            # aiohttp v3.0 has made ClientSession.close a coroutine,
            # so we check whether it is one here and register it
            # to run appropriately at exit
            if asyncio.iscoroutinefunction(self._session.close):
                asyncio.get_event_loop().run_until_complete(
                    self._session.close()
                )
            else:
                self._session.close()

    async def session(self):
        """Returns the underlying `aiohttp.ClientSession`."""
        if isinstance(self._session, self.__ARG_SPEC):
            args, kwargs = self._session
            self._session = aiohttp.ClientSession(*args, **kwargs)
            self._auto_created_session = True
        return self._session

    def wrap_callback(self, callback):
        func = inspect.unwrap(callback)
        if not asyncio.iscoroutinefunction(func):
            callback = self._sync_callback_adapter(callback)
        return callback

    @staticmethod
    @register.handler
    def with_session(session, *args, **kwargs):
        """
        Builds a client instance if the first argument is a
        :py:class:`aiohttp.ClientSession`. Otherwise, return :py:obj:`None`.
        """
        if isinstance(session, aiohttp.ClientSession):
            return AiohttpClient(session, *args, **kwargs)

    @classmethod
    def _create_session(cls, *args, **kwargs):
        return cls.__ARG_SPEC(args, kwargs)

    @classmethod
    def create(cls, *args, **kwargs):
        """
        Builds a client instance with
        :py:class:`aiohttp.ClientSession` arguments.

        Instead of directly initializing this class with a
        :py:class:`aiohttp.ClientSession`, use this method to have the
        client lazily construct a session when sending the first
        request. Hence, this method guarantees that the creation of the
        underlying session happens inside of a coroutine.

        Args:
            *args: positional arguments that
                :py:class:`aiohttp.ClientSession` takes.
            **kwargs: keyword arguments that
                :py:class:`aiohttp.ClientSession` takes.
        """
        session_build_args = cls._create_session(*args, **kwargs)
        return AiohttpClient(session=session_build_args)

    async def send(self, request):
        method, url, extras = request
        session = await self.session()
        response = await session.request(method, url, **extras)

        # Make `aiohttp` response "quack" like a `requests` response
        response.status_code = response.status

        return response

    def apply_callback(self, callback, response):
        return self.wrap_callback(callback)(response)

    @staticmethod
    def io():
        return io.AsyncioStrategy()


class ThreadedCoroutine(object):
    def __init__(self, coroutine):
        self.__coroutine = coroutine

    def __call__(self, *args, **kwargs):
        with AsyncioExecutor() as executor:
            future = executor.submit(self.__coroutine, *args, **kwargs)
            result = future.result()
        return result


class ThreadedResponse(object):
    def __init__(self, response):
        self.__response = response

    def __getattr__(self, item):
        value = getattr(self.__response, item)
        if asyncio.iscoroutinefunction(value):
            return ThreadedCoroutine(value)
        return value

    def unwrap(self):
        return self.__response


class AsyncioExecutor(futures.Executor):
    """
    Executor that runs asyncio coroutines in a shadow thread.

    Credit to Vincent Michel, who wrote the original implementation:
    https://gist.github.com/vxgmichel/d16e66d1107a369877f6ef7e646ac2e5
    """

    def __init__(self):
        self._loop = asyncio.new_event_loop()
        self._thread = threading.Thread(target=self._target)
        self._thread.start()

    def _target(self):
        asyncio.set_event_loop(self._loop)
        self._loop.run_forever()

    def submit(self, fn, *args, **kwargs):
        coro = fn(*args, **kwargs)
        return asyncio.run_coroutine_threadsafe(coro, self._loop)

    def shutdown(self, wait=True):
        self._loop.call_soon_threadsafe(self._loop.stop)
        if wait:  # pragma: no cover
            self._thread.join()


# === Register client exceptions === #
if aiohttp is not None:  # pragma: no cover
    AiohttpClient.exceptions.BaseClientException = aiohttp.ClientError
    AiohttpClient.exceptions.ConnectionError = aiohttp.ClientConnectionError
    AiohttpClient.exceptions.ConnectionTimeout = aiohttp.ClientConnectorError
    AiohttpClient.exceptions.ServerTimeout = aiohttp.ServerTimeoutError
    AiohttpClient.exceptions.SSLError = aiohttp.ClientSSLError
    AiohttpClient.exceptions.InvalidURL = aiohttp.InvalidURL



# Local imports
from uplink import decorators
from uplink.clients.io import RequestTemplate, transitions
from uplink.retry import (
    when as when_mod,
    stop as stop_mod,
    backoff as backoff_mod,
)
from uplink.retry._helpers import ClientExceptionProxy

__all__ = ["retry"]


# noinspection PyPep8Naming
class retry(decorators.MethodAnnotation):
    """
    A decorator that adds retry support to a consumer method or to an
    entire consumer.

    Unless you specify the ``when`` or ``on_exception`` argument, all
    failed requests that raise an exception are retried.

    Unless you specify the ``max_attempts`` or ``stop`` argument, this
    decorator continues retrying until the server returns a response.

    Unless you specify the ``backoff`` argument, this decorator uses
    `capped exponential backoff and jitter <https://amzn.to/2xc2nK2>`_,
    which should benefit performance with remote services under high
    contention.

    .. note::

        Response and error handlers (see :ref:`here <custom response
        handler>`) are invoked after the retry condition breaks or all
        retry attempts are exhausted, whatever comes first. These
        handlers will receive the first response/exception that triggers
        the retry's ``stop`` condition or doesn't match its ``when``
        filter.

        In other words, responses or exceptions that match
        the retry condition (e.g., retry when status code is 5xx) are
        not subject to response or error handlers as long as the request
        doesn't break the retry's stop condition (e.g., stop retrying
        after 5 attempts).

    Args:
        when (optional): A predicate that determines when a retry
            should be attempted.
        max_attempts (int, optional): The number of retries to attempt.
            If not specified, requests are retried continuously until
            a response is rendered.
        on_exception (:class:`Exception`, optional): The exception type
            that should prompt a retry attempt.
        stop (:obj:`callable`, optional): A function that creates
            predicates that decide when to stop retrying a request.
        backoff (:class:`RetryBackoff`, :obj:`callable`, optional): A
            backoff strategy or a function that creates an iterator
            over the ordered sequence of timeouts between retries. If
            not specified, exponential backoff is used.
    """

    _DEFAULT_PREDICATE = when_mod.raises(Exception)

    stop = stop_mod
    backoff = backoff_mod
    when = when_mod

    def __init__(
        self,
        when=None,
        max_attempts=None,
        on_exception=None,
        stop=None,
        backoff=None,
    ):
        if stop is None:
            if max_attempts is not None:
                stop = stop_mod.after_attempt(max_attempts)
            else:
                stop = stop_mod.NEVER

        if on_exception is not None:
            when = when_mod.raises(on_exception) | when

        if when is None:
            when = self._DEFAULT_PREDICATE

        if backoff is None:
            backoff = backoff_mod.jittered()

        if not isinstance(backoff, backoff_mod.RetryBackoff):
            backoff = backoff_mod.from_iterable_factory(backoff)

        self._when = when
        self._backoff = backoff
        self._stop = stop

    BASE_CLIENT_EXCEPTION = ClientExceptionProxy(
        lambda ex: ex.BaseClientException
    )
    CONNECTION_ERROR = ClientExceptionProxy(lambda ex: ex.ConnectionError)
    CONNECTION_TIMEOUT = ClientExceptionProxy(lambda ex: ex.ConnectionTimeout)
    SERVER_TIMEOUT = ClientExceptionProxy(lambda ex: ex.ServerTimeout)
    SSL_ERROR = ClientExceptionProxy(lambda ex: ex.SSLError)

    def modify_request(self, request_builder):
        request_builder.add_request_template(
            _RetryTemplate(
                self._when(request_builder),
                self._backoff,
                self._stop,
            )
        )


class _RetryTemplate(RequestTemplate):
    def __init__(self, condition, backoff, stop):
        self._condition = condition
        self._backoff = backoff
        self._stop = stop
        self._stop_iter = self._stop()

    def _process_timeout(self, timeout):
        next(self._stop_iter)
        if timeout is None or self._stop_iter.send(timeout):
            self._backoff.handle_after_final_retry()
            self._stop_iter = self._stop()
            return None
        return transitions.sleep(timeout)

    def after_response(self, request, response):
        if not self._condition.should_retry_after_response(response):
            return self._process_timeout(None)
        return self._process_timeout(
            self._backoff.get_timeout_after_response(request, response)
        )

    def after_exception(self, request, exc_type, exc_val, exc_tb):
        if not self._condition.should_retry_after_exception(
            exc_type, exc_val, exc_tb
        ):
            return self._process_timeout(None)
        return self._process_timeout(
            self._backoff.get_timeout_after_exception(
                request, exc_type, exc_val, exc_tb
            )
        )



# Third-party imports
import pytest

# Local imports.
from uplink.clients import io

from . import test_retry


@pytest.mark.asyncio
async def test_retry_with_asyncio(mock_client, mock_response):
    import asyncio

    @asyncio.coroutine
    def coroutine():
        return mock_response

    # Setup
    mock_response.with_json({"id": 123, "name": "prkumar"})
    mock_client.with_side_effect([Exception, coroutine()])
    mock_client.with_io(io.AsyncioStrategy())
    github = test_retry.GitHub(base_url=test_retry.BASE_URL, client=mock_client)

    # Run
    response = await github.get_user("prkumar")

    # Verify
    assert len(mock_client.history) == 2
    assert response.json() == {"id": 123, "name": "prkumar"}



# Standard library imports
import contextlib
import math
import threading
import time
import sys

# Local imports
from uplink import decorators, utils
from uplink.clients.io import RequestTemplate, transitions

__all__ = ["ratelimit", "RateLimitExceeded"]

# Use monotonic time if available, otherwise fall back to the system clock.
now = time.monotonic if hasattr(time, "monotonic") else time.time


def _get_host_and_port(base_url):
    parsed_url = utils.urlparse.urlparse(base_url)
    return parsed_url.hostname, parsed_url.port


class RateLimitExceeded(RuntimeError):
    """A request failed because it exceeded the client-side rate limit."""

    def __init__(self, calls, period):
        super(RateLimitExceeded, self).__init__(
            "Exceeded rate limit of [%s] calls every [%s] seconds."
            % (calls, period)
        )


class Limiter(object):
    _last_reset = _num_calls = None

    def __init__(self, max_calls, period, clock):
        self._max_calls = max_calls
        self._period = period
        self._clock = clock
        self._lock = threading.RLock()
        self._reset()

    @property
    def period_remaining(self):
        return self._period - (self._clock() - self._last_reset)

    @contextlib.contextmanager
    def check(self):
        with self._lock:
            if self.period_remaining <= 0:
                self._reset()
            yield self._max_calls > self._num_calls
            self._num_calls += 1

    def _reset(self):
        self._num_calls = 0
        self._last_reset = self._clock()


class RateLimiterTemplate(RequestTemplate):
    def __init__(self, limiter, create_limit_reached_exception):
        self._limiter = limiter
        self._create_limit_reached_exception = create_limit_reached_exception

    def before_request(self, request):
        with self._limiter.check() as ok:
            if ok:
                return  # Fallback to default behavior
            elif self._create_limit_reached_exception is not None:
                raise self._create_limit_reached_exception()
            else:
                return transitions.sleep(self._limiter.period_remaining)


# noinspection PyPep8Naming
class ratelimit(decorators.MethodAnnotation):
    """
    A decorator that constrains a consumer method or an entire
    consumer to making a specified maximum number of requests within a
    defined time period (e.g., 15 calls every 15 minutes).

    Note:
        The rate limit is enforced separately for each host-port
        combination. Logically, requests are grouped by host and port,
        and the number of requests within a time period are counted and
        capped separately for each group.

    By default, when the limit is reached, the client will wait until
    the current period is over before executing any subsequent
    requests. If you'd prefer the client to raise an exception when the
    limit is exceeded, set the ``raise_on_limit`` argument.

    Args:
        calls (int): The maximum number of allowed calls that the
            consumer can make within the time period.
        period (float): The duration of each time period in seconds.
        raise_on_limit (:class:`Exception` or bool, optional): Either an
            exception to raise when the client exceeds the rate limit or
            a :class:`bool`. If :obj:`True`, a
            :class:`~uplink.ratelimit.RateLimitExceeded` exception is
            raised.
    """

    BY_HOST_AND_PORT = _get_host_and_port

    def __init__(
        self,
        calls=15,
        period=900,
        raise_on_limit=False,
        group_by=BY_HOST_AND_PORT,
        clock=now,
    ):
        self._max_calls = max(1, min(sys.maxsize, math.floor(calls)))
        self._period = period
        self._clock = clock
        self._limiter_cache = {}
        self._group_by = utils.no_op if group_by is None else group_by

        if utils.is_subclass(raise_on_limit, Exception) or isinstance(
            raise_on_limit, Exception
        ):
            self._create_limit_reached_exception = raise_on_limit
        elif raise_on_limit:
            self._create_limit_reached_exception = (
                self._create_rate_limit_exceeded
            )
        else:
            self._create_limit_reached_exception = None

    def _get_limiter_for_request(self, request_builder):
        key = self._group_by(request_builder.base_url)
        try:
            return self._limiter_cache[key]
        except KeyError:
            return self._limiter_cache.setdefault(
                key, Limiter(self._max_calls, self._period, self._clock)
            )

    def modify_request(self, request_builder):
        limiter = self._get_limiter_for_request(request_builder)
        request_builder.add_request_template(
            RateLimiterTemplate(limiter, self._create_limit_reached_exception)
        )

    def _create_rate_limit_exceeded(self):
        return RateLimitExceeded(self._max_calls, self._period)



# Standard library imports
import sys
import time

# Local imports
from uplink.clients.io import interfaces

__all__ = ["BlockingStrategy"]


class BlockingStrategy(interfaces.IOStrategy):
    """A blocking execution strategy."""

    def invoke(self, func, arg, kwargs, callback):
        try:
            response = func(*arg, **kwargs)
        except Exception as error:
            tb = sys.exc_info()[2]
            return callback.on_failure(type(error), error, tb)
        else:
            return callback.on_success(response)

    def sleep(self, duration, callback):
        time.sleep(duration)
        return callback.on_success()

    def finish(self, response):
        return response

    def execute(self, executable):
        return executable.execute()



from uplink.clients.io.interfaces import (
    Client,
    Executable,
    IOStrategy,
    RequestTemplate,
)
from uplink.clients.io.execution import RequestExecutionBuilder
from uplink.clients.io.templates import CompositeRequestTemplate
from uplink.clients.io.blocking_strategy import BlockingStrategy

__all__ = [
    "Client",
    "CompositeRequestTemplate",
    "Executable",
    "IOStrategy",
    "RequestTemplate",
    "BlockingStrategy",
    "AsyncioStrategy",
    "TwistedStrategy",
    "RequestExecutionBuilder",
]

try:
    from uplink.clients.io.asyncio_strategy import AsyncioStrategy
except (ImportError, SyntaxError):  # pragma: no cover

    class AsyncioStrategy(IOStrategy):
        def __init__(self, *args, **kwargs):
            raise NotImplementedError(
                "Failed to load `asyncio` execution strategy: you may be using a version "
                "of Python below 3.3. `aiohttp` requires Python 3.4+."
            )


try:
    from uplink.clients.io.twisted_strategy import TwistedStrategy
except (ImportError, SyntaxError):  # pragma: no cover

    class TwistedStrategy(IOStrategy):
        def __init__(self, *args, **kwargs):
            raise NotImplementedError(
                "Failed to load `twisted` execution strategy: you may be not have "
                "the twisted library installed."
            )



# Standard library imports
import inspect
import collections
import pkg_resources


_INSTALLERS = collections.OrderedDict()
_ENTRY_POINTS = collections.OrderedDict()


class plugin(object):
    _BASE_ENTRY_POINT_NAME = "uplink.plugins."

    def __init__(self, name, _entry_points=_ENTRY_POINTS):
        self._name = self._BASE_ENTRY_POINT_NAME + name
        self._entry_points = _entry_points

    def __call__(self, func):
        self._entry_points[self._name] = func
        return func


class installer(object):
    def __init__(self, base_cls, _installers=_INSTALLERS):
        self._base_cls = base_cls
        self._installers = _installers

    def __call__(self, func):
        self._installers[self._base_cls] = func
        return func


def load_entry_points(
    _entry_points=_ENTRY_POINTS,
    _iter_entry_points=pkg_resources.iter_entry_points,
):
    for name in _entry_points:
        plugins = {
            entry_point.name: entry_point.load()
            for entry_point in _iter_entry_points(name)
        }
        func = _entry_points[name]
        for value in plugins.values():
            func(value)


def install(installable, _installers=_INSTALLERS):
    cls = installable if inspect.isclass(installable) else type(installable)
    for base_cls in _installers:
        if issubclass(cls, base_cls):
            _installers[base_cls](installable)
            break
    else:
        raise TypeError("Failed to install: '%s'" % str(installable))

    return installable



# Standard library imports
import os
from setuptools import setup, find_packages


def read(filename):
    with open(filename) as stream:
        return stream.read()


# Read package metadata from __about__.py, to avoid importing the whole
# package prior to installation.
about = dict()
with open(os.path.join("uplink", "__about__.py")) as fp:
    exec(fp.read(), about)
    about = dict((k.strip("_"), about[k]) for k in about)

install_requires = ["requests>=2.18.0", "six>=1.13.0", "uritemplate>=3.0.0"]

extras_require = {
    "marshmallow": ["marshmallow>=2.15.0"],
    "pydantic:python_version >= '3.6'": ["pydantic>=1.6.1"],
    "aiohttp:python_version <= '3.4'": [],
    "aiohttp:python_version >= '3.4'": "aiohttp>=2.3.0",
    "twisted:python_version != '3.3' and python_version != '3.4'": "twisted>=17.1.0",
    # Twisted 18.4.0 dropped py3.3 support
    "twisted:python_version == '3.3'": "twisted<=17.9.0",
    # Twisted 19.7.0 dropped py3.4 support
    "twisted:python_version == '3.4'": "twisted<=19.2.1",
    "typing": ["typing>=3.6.4"],
    "tests": ["pytest", "pytest-mock", "pytest-cov", "pytest-twisted"],
    "tests:python_version >= '3.5'": ["pytest-asyncio"],
}

metadata = {
    "author": "P. Raj Kumar",
    "author_email": "raj.pritvi.kumar@gmail.com",
    "url": "https://uplink.readthedocs.io/",
    "project_urls": {
        "Source": "https://github.com/prkumar/uplink",
    },
    "license": "MIT",
    "description": "A Declarative HTTP Client for Python.",
    "long_description": read("README.rst"),
    "classifiers": [
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    "keywords": "http api rest client retrofit",
    "packages": find_packages(exclude=("tests", "tests.*")),
    "install_requires": install_requires,
    "extras_require": extras_require,
}
metadata = dict(metadata, **about)

if __name__ == "__main__":
    setup(name="uplink", **metadata)



# Third-party imports
import pytest
import requests

# Local imports
from uplink import clients
from tests.integration import MockClient, MockResponse


@pytest.fixture
def mock_client(mocker):
    client = mocker.Mock(spec=clients.interfaces.HttpClientAdapter)
    return MockClient(client)


@pytest.fixture
def mock_response(mocker):
    response = mocker.Mock(spec=requests.Response)
    return MockResponse(response)



# Standard imports
import random
import sys

# Constants
MAX_VALUE = sys.maxsize / 2

__all__ = ["jittered", "exponential", "fixed"]


def from_iterable(iterable):
    """Creates a retry strategy from an iterable of timeouts"""

    class IterableRetryBackoff(_IterableBackoff):
        def __iter__(self):
            return iter(iterable)

    return IterableRetryBackoff()


def from_iterable_factory(iterable_factory):
    """
    Creates a retry strategy from a function that returns an iterable
    of timeouts.
    """

    class IterableRetryBackoff(_IterableBackoff):
        def __iter__(self):
            return iter(iterable_factory())

    return IterableRetryBackoff()


class RetryBackoff(object):
    """
    Base class for a strategy that calculates the timeout between
    retry attempts.

    You can compose two ``RetryBackoff`` instances by using the ``|``
    operator::

        CustomBackoffA() | CustomBackoffB()

    The resulting backoff strategy will first compute the timeout using
    the left-hand instance. If that timeout is ``None``, the strategy
    will try to compute a fallback using the right-hand instance. If
    both instances return ``None``, the resulting strategy will also
    return ``None``.
    """

    def get_timeout_after_response(self, request, response):
        """
        Returns the number of seconds to wait before retrying the
        request, or ``None`` to indicate that the given response should
        be returned.
        """
        raise NotImplementedError  # pragma: no cover

    def get_timeout_after_exception(self, request, exc_type, exc_val, exc_tb):
        """
        Returns the number of seconds to wait before retrying the
        request, or ``None`` to indicate that the given exception
        should be raised.
        """
        raise NotImplementedError  # pragma: no cover

    def handle_after_final_retry(self):
        """
        Handles any clean-up necessary following the final retry
        attempt.
        """
        pass  # pragma: no cover

    def __or__(self, other):
        """Composes the current strategy with another."""
        assert isinstance(
            other, RetryBackoff
        ), "Both objects should be backoffs."
        return _Or(self, other)


class _Or(RetryBackoff):
    def __init__(self, left, right):
        self._left = left
        self._right = right

    def get_timeout_after_response(self, request, response):
        timeout = self._left.get_timeout_after_response(request, response)
        if timeout is None:
            return self._right.get_timeout_after_response(request, response)
        return timeout

    def get_timeout_after_exception(self, request, exc_type, exc_val, exc_tb):
        timeout = self._left.get_timeout_after_exception(
            request, exc_type, exc_val, exc_tb
        )
        if timeout is None:
            return self._right.get_timeout_after_exception(
                request, exc_type, exc_val, exc_tb
            )
        return timeout

    def handle_after_final_retry(self):
        self._left.handle_after_final_retry()
        self._right.handle_after_final_retry()


class _IterableBackoff(RetryBackoff):
    __iterator = None

    def __iter__(self):
        raise NotImplementedError  # pragma: no cover

    def __call__(self):
        return iter(self)

    def __next(self):
        if self.__iterator is None:
            self.__iterator = iter(self)

        try:
            return next(self.__iterator)
        except StopIteration:
            return None

    def get_timeout_after_response(self, request, response):
        return self.__next()

    def get_timeout_after_exception(self, request, exc_type, exc_val, exc_tb):
        return self.__next()

    def handle_after_final_retry(self):
        self.__iterator = None


class jittered(_IterableBackoff):
    """
    Waits using capped exponential backoff and full jitter.

    The implementation is discussed in `this AWS Architecture Blog
    post <https://amzn.to/2xc2nK2>`_, which recommends this approach
    for any remote clients, as it minimizes the total completion
    time of competing clients in a distributed system experiencing
    high contention.
    """

    def __init__(self, base=2, multiplier=1, minimum=0, maximum=MAX_VALUE):
        self._exp_backoff = exponential(base, multiplier, minimum, maximum)

    def __iter__(self):
        for delay in self._exp_backoff():  # pragma: no branch
            yield random.uniform(0, 1) * delay


class exponential(_IterableBackoff):
    """
    Waits using capped exponential backoff, meaning that the delay
    is multiplied by a constant ``base`` after each attempt, up to
    an optional ``maximum`` value.
    """

    def __init__(self, base=2, multiplier=1, minimum=0, maximum=MAX_VALUE):
        self._base = base
        self._multiplier = multiplier
        self._minimum = minimum
        self._maximum = maximum

    def __iter__(self):
        delay = self._multiplier
        while self._minimum > delay:
            delay *= self._base
        while True:
            yield min(delay, self._maximum)
            delay *= self._base


class fixed(_IterableBackoff):
    """Waits for a fixed number of ``seconds`` before each retry."""

    def __init__(self, seconds):
        self._seconds = seconds

    def __iter__(self):
        while True:
            yield self._seconds



# Local imports.
import uplink

# Third-party imports
import aiohttp
import pytest
from uplink.clients.aiohttp_ import AiohttpClient

# Constants
BASE_URL = "https://example.com/"
SIMPLE_RESPONSE = "simple response"


@pytest.fixture
def mock_aiohttp_session(mocker):
    return mocker.Mock(spec=aiohttp.ClientSession)


@uplink.response_handler
async def simple_async_handler(response):
    return SIMPLE_RESPONSE


class Calendar(uplink.Consumer):
    @simple_async_handler
    @uplink.get("todos/{todo_id}")
    def get_todo(self, todo_id):
        pass


@pytest.mark.asyncio
async def test_simple_async_handler(mock_aiohttp_session, mock_response):
    mock_response.status = 200

    async def request(*args, **kwargs):
        return mock_response

    mock_aiohttp_session.request = request

    calendar = Calendar(
        base_url=BASE_URL, client=AiohttpClient(mock_aiohttp_session)
    )

    # Run
    response = await calendar.get_todo(todo_id=1)

    # Verify
    assert response == SIMPLE_RESPONSE



# Local imports
from uplink.retry._helpers import ClientExceptionProxy

__all__ = ["RetryPredicate", "raises", "status", "status_5xx"]


class RetryPredicate(object):
    def should_retry_after_response(self, response):
        return False

    def should_retry_after_exception(self, exc_type, exc_val, exc_tb):
        return False

    def __call__(self, request_builder):  # Pragma: no cover
        return self

    def __or__(self, other):
        if other is not None:
            assert isinstance(
                other, RetryPredicate
            ), "Both objects should be retry conditions."
            return _Or(self, other)
        return self


class _Or(RetryPredicate):
    def __init__(self, left, right):
        self._left = left
        self._right = right

    def should_retry_after_response(self, *args, **kwargs):
        return self._left.should_retry_after_response(
            *args, **kwargs
        ) or self._right.should_retry_after_response(*args, **kwargs)

    def should_retry_after_exception(self, *args, **kwargs):
        return self._left.should_retry_after_exception(
            *args, **kwargs
        ) or self._right.should_retry_after_exception(*args, **kwargs)

    def __call__(self, request_builder):
        left = self._left(request_builder)
        right = self._right(request_builder)
        return type(self)(left, right)


# noinspection PyPep8Naming
class raises(RetryPredicate):
    """Retry when a specific exception type is raised."""

    def __init__(self, expected_exception):
        self._expected_exception = expected_exception

    def __call__(self, request_builder):
        proxy = ClientExceptionProxy.wrap_proxy_if_necessary(
            self._expected_exception
        )
        type_ = proxy(request_builder.client.exceptions)
        return raises(type_)

    def should_retry_after_exception(self, exc_type, exc_val, exc_tb):
        return isinstance(exc_val, self._expected_exception)


# noinspection PyPep8Naming
class status(RetryPredicate):
    """Retry on specific HTTP response status codes."""

    def __init__(self, *status_codes):
        self._status_codes = status_codes

    def should_retry_after_response(self, response):
        return response.status_code in self._status_codes


# noinspection PyPep8Naming
class status_5xx(RetryPredicate):
    """Retry after receiving a 5xx (server error) response."""

    def should_retry_after_response(self, response):
        return 500 <= response.status_code < 600



# Local imports
from uplink import returns


def test_returns(request_builder):
    custom = returns(str)
    request_builder.get_converter.return_value = str
    request_builder.return_type = returns.ReturnType.with_decorator(
        None, custom
    )
    custom.modify_request(request_builder)
    assert request_builder.return_type(2) == "2"


def test_returns_with_multiple_decorators(request_builder, mocker):
    decorator1 = returns(str)
    decorator2 = returns.json()
    request_builder.get_converter.return_value = str
    first_type = returns.ReturnType.with_decorator(None, decorator1)
    second_type = (
        request_builder.return_type
    ) = returns.ReturnType.with_decorator(first_type, decorator2)

    # Verify that the return type doesn't change after being handled by first decorator
    decorator1.modify_request(request_builder)
    assert request_builder.return_type is second_type

    # Verify that the second decorator does handle the return type
    mock_response = mocker.Mock()
    mock_response.json.return_value = {"key": "value"}
    decorator2.modify_request(request_builder)
    assert request_builder.return_type(mock_response) == str(
        mock_response.json()
    )


def test_returns_json(request_builder, mocker):
    mock_response = mocker.Mock()
    mock_response.json.return_value = {"key": "value"}
    request_builder.get_converter.return_value = str
    returns_json = returns.json(str, ())
    request_builder.return_type = returns.ReturnType.with_decorator(
        None, returns_json
    )
    returns_json.modify_request(request_builder)
    assert isinstance(request_builder.return_type, returns.ReturnType)
    assert callable(request_builder.return_type)
    assert request_builder.return_type(mock_response) == str(
        mock_response.json()
    )

    # Verify: Idempotent
    returns_json.modify_request(request_builder)
    assert isinstance(request_builder.return_type, returns.ReturnType)
    assert callable(request_builder.return_type)
    assert request_builder.return_type(mock_response) == str(
        mock_response.json()
    )

    # Verify: Returns JSON when type cannot be converted
    request_builder.get_converter.return_value = None
    returns_json = returns.json(None, ())
    request_builder.return_type = returns.ReturnType.with_decorator(
        None, returns_json
    )
    returns_json.modify_request(request_builder)
    assert callable(request_builder.return_type)
    assert request_builder.return_type(mock_response) == mock_response.json()


def test_returns_json_builtin_type(request_builder, mocker):
    mock_response = mocker.Mock()
    mock_response.json.return_value = {"key": "1"}
    request_builder.get_converter.return_value = None
    returns_json = returns.json(type=int, key="key")
    request_builder.return_type = returns.ReturnType.with_decorator(
        None, returns_json
    )
    returns_json.modify_request(request_builder)
    print(request_builder.return_type)
    assert callable(request_builder.return_type)
    assert request_builder.return_type(mock_response) == 1


class TestReturnsJsonCast(object):
    default_value = {"key": "1"}

    @staticmethod
    def prepare_test(request_builder, mocker, value=default_value, **kwargs):
        mock_response = mocker.Mock()
        mock_response.json.return_value = value
        request_builder.get_converter.return_value = None
        returns_json = returns.json(**kwargs)
        request_builder.return_type = returns.ReturnType.with_decorator(
            None, returns_json
        )
        returns_json.modify_request(request_builder)
        return mock_response

    def test_without_type(self, request_builder, mocker):
        mock_response = self.prepare_test(request_builder, mocker, key="key")
        assert request_builder.return_type(mock_response) == "1"

    def test_with_callable_type(self, request_builder, mocker):
        mock_response = self.prepare_test(
            request_builder,
            mocker,
            type=lambda _: "test",
        )
        assert request_builder.return_type(mock_response) == "test"

    def test_with_builtin_type(self, request_builder, mocker):
        mock_response = self.prepare_test(request_builder, mocker, type=str)
        assert request_builder.return_type(mock_response) == str(
            self.default_value
        )

    def test_with_builtin_type_and_key(self, request_builder, mocker):
        mock_response = self.prepare_test(
            request_builder, mocker, key="key", type=int
        )
        assert request_builder.return_type(mock_response) == 1

    def test_with_not_callable_cast(self, request_builder, mocker):
        mock_response = self.prepare_test(request_builder, mocker, type=1)
        assert request_builder.return_type(mock_response) == self.default_value


def test_returns_JsonStrategy(mocker):
    response = mocker.Mock(spec=["json"])
    response.json.return_value = {"hello": "world"}
    converter = returns.JsonStrategy(lambda x: x, "hello")
    assert converter(response) == "world"

    converter = returns.JsonStrategy(lambda y: y + "!", "hello")
    assert converter(response) == "world!"
